import pytest

from alchemy_cat.dag.core import Graph, PyungoError
from alchemy_cat.dag.io import Input, Output


def test_simple():
    graph = Graph()

    @graph.register(inputs=['a', 'b'], outputs=['c'])
    def f_my_function(a, b):
        return a + b

    @graph.register(inputs=['d', 'a'], outputs=['e'])
    def f_my_function3(d, a):
        return d - a

    @graph.register(inputs=['c'], outputs=['d'])
    def f_my_function2(c):
        return c / 10.

    res = graph.calculate(data={'a': 2, 'b': 3})
    assert res == -1.5
    assert graph.data['e'] == -1.5

    # make sure it is independent
    res = graph.calculate(data={'a': 2, 'b': 3})
    assert res == -1.5
    assert graph.data['e'] == -1.5


def test_call():
    graph = Graph()

    @graph.register(inputs=['a', 'b'], outputs=['c'])
    def f_my_function(a, b):
        return a + b

    @graph.register(inputs=['d', 'a'], outputs=['e'])
    def f_my_function3(d, a):
        return d - a

    @graph.register(inputs=['c'], outputs=['d'])
    def f_my_function2(c):
        return c / 10.

    for _ in range(2):
        res = graph(a=2, b=3)
        assert res == -1.5
        assert graph.data['e'] == -1.5

    with pytest.raises(PyungoError, match="Graph only receive keyword args which will be "
                                          "recognized as input name and value."):
        graph(2, 3)


def test_inputs_args_equivalent():
    graph = Graph()

    @graph.register(args=['a', 'b'], outputs=['c'])
    def f_my_function(a, b):
        return a + b

    @graph.register(args=['d', 'a'], outputs=['e'])
    def f_my_function3(d, a):
        return d - a

    @graph.register(args=['c'], outputs=['d'])
    def f_my_function2(c):
        return c / 10.

    res = graph.calculate(data={'a': 2, 'b': 3})
    assert res == -1.5
    assert graph.data['e'] == -1.5

    # make sure it is independent
    res = graph.calculate(data={'a': 2, 'b': 3})
    assert res == -1.5
    assert graph.data['e'] == -1.5


def test_inputs_args_blend():
    graph = Graph()

    @graph.register(inputs=['a'], args=['b'], outputs=['c'])
    def f_my_function(a, b):
        return a + b

    @graph.register(inputs=['d'], args=['a'], outputs=['e'])
    def f_my_function3(d, a):
        return d - a

    @graph.register(args=['c'], outputs=['d'])
    def f_my_function2(c):
        return c / 10.

    res = graph.calculate(data={'a': 2, 'b': 3})
    assert res == -1.5
    assert graph.data['e'] == -1.5

    # make sure it is independent
    res = graph.calculate(data={'a': 2, 'b': 3})
    assert res == -1.5
    assert graph.data['e'] == -1.5


def test_simple_constant_inputs():
    graph = Graph()

    @graph.register(inputs=[{'a': 2}, {'b': 3}], outputs=['c'])
    def f_my_function(a, b):
        return a + b

    @graph.register(inputs=['d', 'a'], outputs=['e'])
    def f_my_function3(d, a):
        return d - a

    @graph.register(inputs=['c'], outputs=['d'])
    def f_my_function2(c):
        return c / 10.

    res = graph.calculate(data={'a': 2})
    assert res == -1.5
    assert graph.data['e'] == -1.5

    # make sure it is independent
    res = graph.calculate(data={'a': 2})
    assert res == -1.5
    assert graph.data['e'] == -1.5


def test_constant_args_kwargs():
    graph = Graph()

    @graph.register(
        inputs=['a', 'b'],
        args=['c', {'cc': 6}],
        kwargs=['d', {'dc': 7}],
        outputs=['e']
    )
    def f_my_function(a, b, *args, **kwargs):
        return a + b + args[0] + args[1] + kwargs['d'] + kwargs['dc']

    res = graph.calculate(data={'a': 2, 'b': 3, 'c': 4, 'd': 5})

    assert res == 27
    assert graph.data['e'] == 27


def test_constant_dict_kwargs():
    graph = Graph()

    @graph.register(
        inputs=['a', 'b'],
        args=['c'],
        kwargs={'d': 5, 'dc': 6},
        outputs=['e']
    )
    def f_my_function(a, b, *args, **kwargs):
        return a + b + args[0] + kwargs['d'] + kwargs['dc']

    res = graph.calculate(data={'a': 2, 'b': 3, 'c': 4})

    assert res == 20
    assert graph.data['e'] == 20


def test_complex_constant_inputs():
    """Test complex constant inputs.
        * Test {k:v, k:v, ...} constant
        * Test Input(name, value=*) constant
        * Test kwargs constant
    """
    graph = Graph()

    # f1 = -1
    @graph.register(inputs={'inp_1_1': 2, 'inp_1_2': 3}, kwargs={'inp_1_3': 6}, outputs=['f1'])
    def f_my_function1(inp_1_1, inp_1_2=2, inp_1_3=3):
        return inp_1_1 + inp_1_2 - inp_1_3

    # f2 = (2, -5)
    @graph.register(args=['f1', Input('i_2_2', value=-1), {'i_2_3_1': 1}, {'i_2_3_2': -2}],
                    kwargs=[Input('inp_2_4', value=3)], outputs=['f2'])
    def f_my_function2(inp_2_1=4, inp_2_2=5, *inp_2_3, **inp_2_4):
        return inp_2_1 * inp_2_2 + inp_2_3[0], inp_2_3[1] - list(inp_2_4.values())[0]

    # f3 = -33
    @graph.register(inputs=['f1', 'f2', 'inp_3_3'], outputs=['f3'])
    def f_my_function3(inp_3_1, inp_3_2, inp_3_3=4):
        return (inp_3_1 - inp_3_2[0] * inp_3_2[1]) * inp_3_3

    for _ in range(2):
        res = graph.calculate(data={'inp_3_3': 3})
        assert res == 27
        assert graph.data['f3'] == res
        assert graph.data['f1'] == -1
        assert graph.data['f2'] == (2, -5)


def test_slim_graph():
    import numpy as np
    graph = Graph(slim=True)

    @graph.register(inputs=['a', 'b'], outputs=['c'])
    def f_my_function(a, b):
        return a + b

    @graph.register(inputs=['d', 'a'], outputs=['e'])
    def f_my_function3(d, a):
        return d - a

    @graph.register(inputs=['c'], outputs=['d'])
    def f_my_function2(c):
        return c / 10.

    for _ in range(2):
        data = {'a': np.ones((2, 2)) * 2., 'b': np.ones((2, 2)) * 3.}
        res = graph.calculate(data)
        assert (res == np.ones((2, 2)) * -1.5).all()
        assert (graph.data['e'] == np.ones((2, 2)) * -1.5).all()
        assert id(graph.ordered_nodes[0]._outputs[0].value) == id(graph.ordered_nodes[1]._inputs[0].value)
        assert id(graph.ordered_nodes[1]._outputs[0].value) == id(graph.ordered_nodes[2]._inputs[0].value)
        assert id(data['a']) == id(graph.ordered_nodes[0]._inputs[0].value)
        assert id(data['b']) == id(graph.ordered_nodes[0]._inputs[1].value)


def test_not_slim_graph():
    import numpy as np
    graph = Graph(slim=False)

    @graph.register(inputs=['a', 'b'], outputs=['c'])
    def f_my_function(a, b):
        return a + b

    @graph.register(inputs=['d', 'a'], outputs=['e'])
    def f_my_function3(d, a):
        return d - a

    @graph.register(inputs=['c'], outputs=['d'])
    def f_my_function2(c):
        return c / 10.

    for _ in range(2):
        data = {'a': np.ones((2, 2)) * 2., 'b': np.ones((2, 2)) * 3.}
        res = graph.calculate(data)
        assert (res == np.ones((2, 2)) * -1.5).all()
        assert (graph.data['e'] == np.ones((2, 2)) * -1.5).all()
        assert id(graph.ordered_nodes[0]._outputs[0].value) != id(graph.ordered_nodes[1]._inputs[0].value)
        assert id(graph.ordered_nodes[1]._outputs[0].value) != id(graph.ordered_nodes[2]._inputs[0].value)
        assert id(data['a']) != id(graph.ordered_nodes[0]._inputs[0].value)
        assert id(data['b']) != id(graph.ordered_nodes[0]._inputs[1].value)


def test_node_slim_graph():
    import numpy as np
    graph = Graph(slim=False)

    @graph.register(inputs=['a', 'b'], outputs=['c'])
    def f_my_function(a, b):
        return a + b

    @graph.register(inputs=['d', 'a'], outputs=['e'])
    def f_my_function3(d, a):
        return d - a

    @graph.register(inputs=['c'], outputs=['d'], slim_names=['c'])
    def f_my_function2(c):
        return c / 10.

    res = graph.calculate(data={'a': np.ones((2, 2)) * 2., 'b': np.ones((2, 2)) * 3.})
    assert (res == np.ones((2, 2)) * -1.5).all()
    assert (graph.data['e'] == np.ones((2, 2)) * -1.5).all()
    assert id(graph.ordered_nodes[0]._outputs[0].value) == id(graph.ordered_nodes[1]._inputs[0].value)
    assert id(graph.ordered_nodes[1]._outputs[0].value) != id(graph.ordered_nodes[2]._inputs[0].value)

    res = graph.calculate(data={'a': np.ones((2, 2)) * 2., 'b': np.ones((2, 2)) * 3.})
    assert (res == np.ones((2, 2)) * -1.5).all()
    assert (graph.data['e'] == np.ones((2, 2)) * -1.5).all()
    assert id(graph.ordered_nodes[0]._outputs[0].value) == id(graph.ordered_nodes[1]._inputs[0].value)
    assert id(graph.ordered_nodes[1]._outputs[0].value) != id(graph.ordered_nodes[2]._inputs[0].value)


def test_functor():
    graph = Graph()

    @graph.register(inputs=['a', 'b'], outputs=['c'], init={})
    class f_my_function(object):
        factor = 1

        def __init__(self):
            self.dummy = 100

        def __call__(self, a1, a2):
            return a1 + a2 + self.factor

    @graph.register(inputs=['d', 'a'], outputs=['e'], init={'add_or_sub': 'add', 'factor': 2})
    class f_my_function2(object):
        def __init__(self, add_or_sub='add', factor=1):
            if add_or_sub == 'add':
                self.func = lambda x, y: x + y
            else:
                self.func = lambda x, y: x - y

            self.factor = factor

        def generate_constant(self):
            return 1

        def __call__(self, a1, a2):
            return (self.func(a1, a2) + self.generate_constant()) * self.factor

    @graph.register(inputs=['c', 'b'], outputs=['d'], init={'add_or_sub': 'sub'})
    class f_my_function1(object):
        def __init__(self, add_or_sub='add', factor=1):
            if add_or_sub == 'add':
                self.func = lambda x, y: x + y
            else:
                self.func = lambda x, y: x - y

            self.factor = factor

        def generate_constant(self):
            return 1

        def __call__(self, a1, a2):
            return (self.func(a1, a2) + self.generate_constant()) * self.factor

    res = graph.calculate(data={'a': 2, 'b': 3})
    assert graph.data['c'] == 6
    assert graph.data['d'] == 4
    assert res == 14
    assert graph.data['e'] == 14

    # make sure it is independent
    res = graph.calculate(data={'a': 2, 'b': 3})
    assert graph.data['c'] == 6
    assert graph.data['d'] == 4
    assert res == 14
    assert graph.data['e'] == 14


def test_simple_without_decorator():
    graph = Graph()

    def f_my_function(a, b):
        return a + b

    def f_my_function3(d, a):
        return d - a

    def f_my_function2(c):
        return c / 10.

    graph.add_node(f_my_function, inputs=['a', 'b'], outputs=['c'])
    graph.add_node(f_my_function3, inputs=['d', 'a'], outputs=['e'])
    graph.add_node(f_my_function2, inputs=['c'], outputs=['d'])

    res = graph.calculate(data={'a': 2, 'b': 3})

    assert res == -1.5
    assert graph.data['e'] == -1.5


def par_f_my_function(a, b):
    return a + b


def par_f_my_function3(d, a):
    return d - a


def par_f_my_function2(c):
    return c / 10.


def test_simple_parallel():
    """ TODO: We could mock and make sure things are called correctly """

    graph = Graph(pool_size=2)

    graph.add_node(par_f_my_function, inputs=['a', 'b'], outputs=['c'])
    graph.add_node(par_f_my_function3, inputs=['d', 'a'], outputs=['e'])
    graph.add_node(par_f_my_function2, inputs=['c'], outputs=['d'])
    graph.add_node(par_f_my_function2, inputs=['c'], outputs=['f'])
    graph.add_node(par_f_my_function2, inputs=['c'], outputs=['g'])

    res = graph.calculate(data={'a': 2, 'b': 3})

    assert res == -1.5


def test_multiple_outputs():
    graph = Graph()

    @graph.register(inputs=['a', 'b'], outputs=['c', 'd'])
    def f_my_function(a, b):
        return a + b, 2 * b

    @graph.register(inputs=['c', 'd'], outputs=['e'])
    def f_my_function2(c, d):
        return c + d

    res = graph.calculate(data={'a': 2, 'b': 3})

    assert res == 11
    assert graph.data['e'] == 11


def test_same_output_names():
    graph = Graph()

    @graph.register(inputs=['a', 'b'], outputs=['c'])
    def f_my_function(a, b):
        return a + b

    with pytest.raises(PyungoError) as err:
        @graph.register(inputs=['c'], outputs=['c'])
        def f_my_function2(c):
            return c / 10

    assert "Node Node(<f_my_function2>, ['c'], ['c']) have repeated output names: ['c']" in str(err.value)


def test_missing_input():
    graph = Graph()

    @graph.register(inputs=['a', 'b'], outputs=['c'])
    def f_my_function(a, b):
        return a + b

    with pytest.raises(PyungoError) as err:
        graph.calculate(data={'a': 6})

    assert "The following inputs are needed: ['b']" in str(err.value)


def test_missing_kwargs():
    graph = Graph()

    @graph.register(inputs=['a'], kwargs=['b'], outputs=['c'])
    def f_my_function(a, b):
        return a + b

    with pytest.raises(PyungoError) as err:
        graph.calculate(data={'a': 6})

    assert "The following inputs are needed: ['b']" in str(err.value)


def test_missing_input_both_nec_opt():
    graph = Graph()

    @graph.register(inputs=['a', 'b'], outputs=['c'])
    def f_my_function(a, b=2):
        return a + b

    @graph.register(kwargs=['a', 'b'], outputs=['e'])
    def f_my_function3(a, b):
        return a - b

    @graph.register(inputs=['c', 'e'], outputs=['f'])
    def f_my_function2(c, e):
        return c + e / 10.

    with pytest.raises(PyungoError) as err:
        graph.calculate(data={'a': 6})

    assert "The following inputs are needed: ['b']" in str(err.value)


def test_inputs_not_used():
    graph = Graph()

    @graph.register(inputs=['a', 'b'], outputs=['c'])
    def f_my_function(a, b):
        return a + b

    with pytest.raises(PyungoError) as err:
        graph.calculate(data={'a': 6, 'b': 4, 'e': 7})

    assert "The following inputs are not used by the model: ['e']" in str(err.value)


def test_inputs_not_used_with_constant():
    graph = Graph()

    @graph.register(inputs=[{'a': 1}, 'b'], outputs=['c'])
    def f_my_function(a, b):
        return a + b

    with pytest.raises(PyungoError) as err:
        graph.calculate(data={'a': 6, 'b': 4})

    assert "The following inputs are not used by the model: ['a']" in str(err.value)


def test_opt_inputs_wont_cause_redundant_input():
    graph = Graph()

    @graph.register(inputs=['a', 'b'], outputs=['c'])
    def f_my_function(a, b=2):
        return a + b

    res = graph.calculate(data={'a': 6})
    assert res == 8


def test_inputs_collision():
    graph = Graph()

    @graph.register(inputs=['a', 'b'], outputs=['c'])
    def f_my_function(a, b):
        return a + b

    with pytest.raises(PyungoError) as err:
        graph.calculate(data={'a': 6, 'b': 4, 'c': 7})

    assert "The following inputs are already used in the model: ['c']" in str(err.value)


def test_self_dependence():
    graph = Graph()

    @graph.register(inputs=['a', 'b'], outputs=['c'])
    def f_my_function(a, b):
        return a + b

    with pytest.raises(PyungoError) as err:
        @graph.register(inputs=['c', 'd', 'e', {'f': 1}], outputs=['d', 'e', 'f'])
        def f_my_function2(c, d, e):
            return c, d, e

    assert "Node Node(<f_my_function2>, ['c', 'd', 'e', 'f'], ['d', 'e', 'f']) have self dependence caused " \
           "by the following inputs: ['d', 'e']" in str(err.value)


def test_circular_dependency():
    graph = Graph()

    @graph.register(inputs=['a', 'b', 'd'], outputs=['c'])
    def f_my_function(a, b):
        return a + b

    @graph.register(inputs=['c'], outputs=['d'])
    def f_my_function2(c):
        return c / 2.

    with pytest.raises(PyungoError) as err:
        graph.calculate(data={'a': 6, 'b': 4})

    assert "A cyclic dependency exists amongst" in str(err.value)


def test_iterable_on_single_output():
    graph = Graph()

    @graph.register(inputs=['a', 'b'], outputs=['c'])
    def f_my_function(a, b):
        return list(range(a)) + [b]

    res = graph.calculate(data={'a': 2, 'b': 3})

    assert res == [0, 1, 3]
    assert graph.data['c'] == [0, 1, 3]


def test_multiple_outputs_with_iterable():
    graph = Graph()

    @graph.register(inputs=['a', 'b'], outputs=['c', 'd'])
    def f_my_function(a, b):
        return list(range(a)) + [b], b * 10

    res = graph.calculate(data={'a': 2, 'b': 3})

    assert isinstance(res, tuple) is True
    assert graph.data['c'] == [0, 1, 3]
    assert graph.data['d'] == 30
    assert res[0] == [0, 1, 3]
    assert res[1] == 30


def test_args_kwargs():
    graph = Graph()

    @graph.register(
        inputs=['a', 'b'],
        args=['c'],
        kwargs=['d'],
        outputs=['e']
    )
    def f_my_function(a, b, *args, **kwargs):
        return a + b + args[0] + kwargs['d']

    res = graph.calculate(data={'a': 2, 'b': 3, 'c': 4, 'd': 5})

    assert res == 14
    assert graph.data['e'] == 14


def test_diff_input_function_arg_name():
    graph = Graph()

    @graph.register(
        inputs=['a_diff', 'b_diff'],
        args=['c_diff'],
        kwargs=['d'],
        outputs=['e_diff']
    )
    def f_my_function(a, b, *args, **kwargs):
        return a + b + args[0] + kwargs['d']

    res = graph.calculate(data={'a_diff': 2, 'b_diff': 3, 'c_diff': 4, 'd': 5})

    assert res == 14
    assert graph.data['e_diff'] == 14


def test_dag_pretty_print():
    graph = Graph()

    @graph.register(inputs=['a', 'b'], outputs=['c'])
    def f_my_function(a, b):
        return a + b

    @graph.register(inputs=['d', 'a'], outputs=['e'])
    def f_my_function3(d, a):
        return d - a

    @graph.register(inputs=['c'], outputs=['d'])
    def f_my_function2(c):
        return c / 10.

    expected = ['f_my_function', 'f_my_function2', 'f_my_function3']
    dag = graph.dag
    for i, fct_name in enumerate(expected):
        assert dag[i][0].fct_name == fct_name


def test_passing_data_to_node_definition():
    graph = Graph()

    @graph.register(inputs=['a', {'b': 2}], outputs=['c'])
    def f_my_function(a, b):
        return a + b

    res = graph.calculate(data={'a': 5})
    assert res == 7


def test_Input_type_input():
    graph = Graph()

    @graph.register(
        inputs=[Input(name='a'), Input(name='inp_1_1', map='b')],
        outputs=['c']
    )
    def f_my_function(a, b):
        return a + b

    res = graph.calculate(data={'a': 2, 'b': 3})

    assert res == 5


def test_input_type_tuple():
    graph = Graph()

    @graph.register(
        inputs=[('inp_1', 'a'), ('b', 'b')],
        outputs=['c']
    )
    def f_my_function(a, b):
        return a + b

    res = graph.calculate(data={'a': 2, 'b': 3})

    assert res == 5


def test_wrong_input_type():
    graph = Graph()

    with pytest.raises(PyungoError) as err:
        @graph.register(inputs=['a', {'b'}], outputs=['c'])
        def f_my_function(a, b):
            return a + b

    assert "inputs need to be of type tuple, Input, str or dict" in str(err.value)


def test_input_tuple_too_long():
    graph = Graph()

    with pytest.raises(PyungoError) as err:
        @graph.register(inputs=[('a', 'input', 'too_long'), 'b'], outputs=['c'])
        def f_my_function(a, b):
            return a + b

    assert "Tuple input should like (name, map). However, get ('a', 'input', 'too_long')" in str(err.value)


def test_empty_input_dict():
    graph = Graph()

    with pytest.raises(PyungoError) as err:
        @graph.register(inputs=['a', {}], outputs=['c'])
        def f_my_function(a, b):
            return a + b

    assert "dict inputs should have only one key and cannot be empty" in str(err.value)


def test_multiple_keys_input_dict():
    graph = Graph()

    with pytest.raises(PyungoError) as err:
        @graph.register(inputs=['a', {'b': 1, 'c': 2}], outputs=['c'])
        def f_my_function(a, b):
            return a + b

    assert "dict inputs should have only one key and cannot be empty" in str(err.value)


def test_not_str_name():
    graph = Graph()

    with pytest.raises(PyungoError) as err:
        @graph.register(inputs=[(23, 'a')], outputs=['c'])
        def f_my_function(a, b):
            return a + b

    assert "IO name must be str, however get name = 23 with type <class 'int'>" in str(err.value)


def test_not_str_map():
    graph = Graph()

    with pytest.raises(PyungoError) as err:
        @graph.register(inputs=[Input('a', map=23)], outputs=['c'])
        def f_my_function(a, b):
            return a + b

    assert "IO map must be str, however get map = 23 with type <class 'int'>" in str(err.value)


@pytest.mark.skip("Don't Support Contract Now")
def test_contract_inputs():
    from contracts import ContractNotRespected
    graph = Graph()

    @graph.register(
        inputs=[Input(name='a', contract='int,>0'), 'b'],
        outputs=['c']
    )
    def f_my_function(a, b):
        return a + b

    res = graph.calculate(data={'a': 2, 'b': 3})
    assert res == 5
    res = graph.calculate(data={'a': 2, 'b': 3})
    assert res == 5
    with pytest.raises(ContractNotRespected) as err:
        res = graph.calculate(data={'a': -2, 'b': 3})
    assert "Condition -2 > 0 not respected" in str(err.value)


@pytest.mark.skip("Don't Support Contract Now")
def test_contract_outputs():
    from contracts import ContractNotRespected
    graph = Graph()

    @graph.register(
        inputs=['a', 'b'],
        outputs=[Output('c', contract='int,>0')]
    )
    def f_my_function(a, b):
        return a + b

    res = graph.calculate(data={'a': 2, 'b': 3})
    assert res == 5
    with pytest.raises(ContractNotRespected) as err:
        res = graph.calculate(data={'a': -4, 'b': 3})
    assert "Condition -1 > 0 not respected" in str(err.value)


def test_map():
    graph = Graph()

    @graph.register(
        inputs=[Input('a', map='q'), Input('b', map='w')],
        outputs=[Output('c', map='e')]
    )
    def f_my_function(a, b):
        return a + b

    res = graph.calculate(data={'q': 2, 'w': 3})
    assert res == 5
    assert graph.data['e'] == 5


def test_build_with_map_feed_with_name():
    graph = Graph()

    @graph.register(inputs=[('foo', 'a')], kwargs=[('inp1_2', 'b')], outputs=['c'])
    def f_my_function1(inp1_1, inp1_2):
        return inp1_1 + inp1_2

    @graph.register(args=[Input(name='foo', map='d')], kwargs=[Input(name='inp3_2', map='a')], outputs=['e'])
    def f_my_function3(inp3_1, inp3_2):
        return inp3_1 - inp3_2

    @graph.register(inputs=[('foo', 'c')], outputs=['d'])
    def f_my_function2(inp2_1):
        return inp2_1 / 10.

    res = graph.calculate(data={'a': 2, 'b': 3})
    assert res == -1.5
    assert graph.data['e'] == -1.5

    # make sure it is independent
    res = graph.calculate(data={'a': 2, 'b': 3})
    assert res == -1.5
    assert graph.data['e'] == -1.5


def test_schema():
    from jsonschema import ValidationError

    schema = {
        "type": "object",
        "properties": {
            "a": {"type": "number"},
            "b": {"type": "number"}
        }
    }

    graph = Graph(schema=schema)

    @graph.register(
        inputs=['a', 'b'],
        outputs=['c']
    )
    def f_my_function(a, b):
        return a + b

    with pytest.raises(ValidationError) as err:
        graph.calculate(data={'a': 1, 'b': '2'})

    msg = "'2' is not of type 'number'"
    assert msg in str(err.value)

    res = graph.calculate(data={'a': 1, 'b': 2})
    assert res == 3


def test_find_default_by_name_not_map():
    graph = Graph()

    @graph.register(inputs=['a', ('inp2', 'b')], kwargs=[('inp3', 'c')], outputs=['d'])
    def f(inp1, inp2=2, inp3=3):
        return inp1 + inp2 + inp3

    res = graph.calculate(data={'a': 1})

    assert res == 6
    assert graph.data['d'] == 6


def test_optional_inputs_without_feed():
    graph = Graph()

    @graph.register(inputs=['a', 'b'], kwargs=['c'], outputs=['d'])
    def f(a, b=2, c=3):
        return a + b + c

    res = graph.calculate(data={'a': 1})

    assert res == 6
    assert graph.data['d'] == 6


def test_optional_inputs_feed_by_input():
    graph = Graph()

    @graph.register(inputs=['a', 'b'], kwargs=['c'], outputs=['d'])
    def f(a, b=2, c=3):
        return a + b + c

    res = graph.calculate(data={'a': 1, 'b': -1, 'c': -2})

    assert res == -2
    assert graph.data['d'] == -2


def test_optional_inputs_feed_by_output():
    graph = Graph()

    @graph.register(inputs=['a'], kwargs=['b'], outputs=['c'])
    def f(a, b):
        return a + b

    @graph.register(inputs=['a'], kwargs=['b'], outputs=['d'])
    def f2(a, b):
        return a - b

    @graph.register(inputs=['c'], kwargs=[Input(map='d', name='inp2')], outputs=['e'])
    def f1(inp1=0, inp2=0):
        return inp1 + inp2

    res = graph.calculate(data={'a': 1, 'b': 3})

    assert res == 2
    assert graph.data['e'] == res
    assert graph.data['c'] == 4
    assert graph.data['d'] == -2


def test_no_explicit_inputs_outputs_simple():
    graph = Graph()

    @graph.register()
    def f(a, b):
        c = a + b
        return c

    res = graph.calculate(data={'a': 1, 'b': 2})

    assert res == 3
    assert graph.data['c'] == 3


def test_no_explicit_inputs_outputs_tuple():
    graph = Graph()

    @graph.register()
    def f(a, b, c, d):
        e = a + b
        f = c - d
        return e, f

    res = graph.calculate(data={'a': 1, 'b': 2, 'c': 3, 'd': 4})

    assert res == (3, -1)
    assert graph.data['e'] == 3
    assert graph.data['f'] == -1


def test_no_explicit_inputs_outputs_bad_return():
    graph = Graph()

    with pytest.raises(PyungoError) as err:
        @graph.register()
        def f(a, b):
            return a + b

    expected = ('Variable name or Tuple of variable '
                'names are expected, got BinOp')
    assert str(err.value) == expected


def test_sub_graph():
    def f_my_function1(inp1_1, inp1_2):
        return inp1_1 + inp1_2

    def f_my_function2(inp2_1):
        return inp2_1 / 10.

    def f_my_function3(inp3_1, inp3_2):
        return inp3_1 - inp3_2

    graph0 = Graph()
    graph0_0 = Graph()
    graph0_1 = Graph()
    graph0_1_0 = Graph()

    # inp2_1 = 1.5, inp3_2 = 0.15, out=0.1785
    graph0.add_node(graph0_0, kwargs=['inp1_1', 'inp1_2'], outputs=['inp2_1'])
    graph0.add_node(f_my_function2, inputs=['inp2_1'], outputs=['inp3_2'])
    graph0.add_node(graph0_1, kwargs=[('inp1', 'inp1_1'), ('inp2', 'inp3_2')], outputs=['out'])

    graph0_0.add_node(f_my_function1, inputs=['inp1_1', 'inp1_2'], outputs=['inp2_1'])
    graph0_0.add_node(f_my_function2, inputs=['inp2_1'], outputs=['inp3_2'])
    graph0_0.add_node(f_my_function3, inputs=[('inp3_1', 'inp1_1'), 'inp3_2'], outputs=['out'])

    # inp2_1 = 0.215, inp3_2 = 0.0215, out = 0.1785
    graph0_1_0.add_node(f_my_function1, inputs=['inp1_1', 'inp1_2'], outputs=['inp2_1'])
    graph0_1_0.add_node(f_my_function2, inputs=['inp2_1'], outputs=['inp3_2'])
    graph0_1_0.add_node(f_my_function3, inputs=[('inp3_1', 'inp1_1'), 'inp3_2'], outputs=['out'])

    # out1 = 0.2, out2 = 0.015, out=0.1785
    graph0_1.add_node(f_my_function2, inputs=['inp1'], outputs=['out1'])
    graph0_1.add_node(f_my_function2, inputs=['inp2'], outputs=['out2'])
    graph0_1.add_node(graph0_1_0, kwargs=[('inp1_1', 'out1'), ('inp1_2', 'out2')], outputs=['out3'])

    for _ in range(2):
        res = graph0(inp1_1=2, inp1_2=3)
        assert 0.1785 == pytest.approx(res)
        assert graph0.data['out'] == res
        assert graph0_0.data['out'] == 1.5
        assert graph0_1.data['out3'] == pytest.approx(0.1785)
        assert graph0_1_0.data['out'] == pytest.approx(0.1785)


def test_sub_graph_with_arg_input():
    graph0 = Graph()
    graph0_0 = Graph()

    with pytest.raises(PyungoError) as err:
        graph0.add_node(graph0_0, args=['inp1_1', 'inp1_2'], outputs=['inp2_1'])

    assert "Node with Graph can only accept kwargs input. However, get args = ['inp1_1', 'inp1_2']" in str(err.value)


def test_deep_ordered_nodes():
    def f_my_function1(inp1_1, inp1_2):
        return inp1_1 + inp1_2

    def f_my_function2(inp2_1):
        return inp2_1 / 10.

    def f_my_function3(inp3_1, inp3_2):
        return inp3_1 - inp3_2

    graph0 = Graph()
    graph0_0 = Graph()
    graph0_1 = Graph()
    graph0_1_0 = Graph()

    # inp2_1 = 1.5, inp3_2 = 0.15, out=0.1785
    graph0.add_node(graph0_0, kwargs=['inp1_1', 'inp1_2'], outputs=['inp2_1'])
    graph0.add_node(f_my_function2, inputs=['inp2_1'], outputs=['inp3_2'])
    graph0.add_node(graph0_1, kwargs=[('inp1', 'inp1_1'), ('inp2', 'inp3_2')], outputs=['out'])

    graph0_0.add_node(f_my_function1, inputs=['inp1_1', 'inp1_2'], outputs=['inp2_1'])
    graph0_0.add_node(f_my_function2, inputs=['inp2_1'], outputs=['inp3_2'])
    graph0_0.add_node(f_my_function3, inputs=[('inp3_1', 'inp1_1'), 'inp3_2'], outputs=['out'])

    # inp2_1 = 0.215, inp3_2 = 0.0215, out = 0.1785
    graph0_1_0.add_node(f_my_function1, inputs=['inp1_1', 'inp1_2'], outputs=['inp2_1'])
    graph0_1_0.add_node(f_my_function2, inputs=['inp2_1'], outputs=['inp3_2'])
    graph0_1_0.add_node(f_my_function3, inputs=[('inp3_1', 'inp1_1'), 'inp3_2'], outputs=['out'])

    # out1 = 0.2, out2 = 0.015, out=0.1785
    graph0_1.add_node(f_my_function2, inputs=['inp1'], outputs=['out1'])
    graph0_1.add_node(f_my_function2, inputs=['inp2'], outputs=['out2'])
    graph0_1.add_node(graph0_1_0, kwargs=[('inp1_1', 'out1'), ('inp1_2', 'out2')], outputs=['out3'])

    deep_ordered_nodes = graph0.deep_ordered_nodes
    graph0_nodes = graph0.ordered_nodes
    graph0_0_nodes = graph0_0.ordered_nodes
    graph0_1_nodes = graph0_1.ordered_nodes
    graph0_1_0_nodes = graph0_1_0.ordered_nodes

    assert deep_ordered_nodes[0] is graph0_nodes[0]
    for n1, n2 in zip(deep_ordered_nodes[1:4], graph0_0_nodes):
        assert n1 is n2
    assert deep_ordered_nodes[4] is graph0_nodes[1]
    assert deep_ordered_nodes[5] is graph0_nodes[2]
    for n1, n2 in zip(deep_ordered_nodes[6:9], graph0_1_nodes):
        assert n1 is n2
    for n1, n2 in zip(deep_ordered_nodes[9:], graph0_1_0_nodes):
        assert n1 is n2


def test_deep_prefix_id_ordered_nodes():
    def f_my_function1(inp1_1, inp1_2):
        return inp1_1 + inp1_2

    def f_my_function2(inp2_1):
        return inp2_1 / 10.

    def f_my_function3(inp3_1, inp3_2):
        return inp3_1 - inp3_2

    graph0 = Graph()
    graph0_0 = Graph()
    graph0_1 = Graph()
    graph0_1_0 = Graph()

    # inp2_1 = 1.5, inp3_2 = 0.15, out=0.1785
    graph0.add_node(graph0_0, kwargs=['inp1_1', 'inp1_2'], outputs=['inp2_1'])
    graph0.add_node(f_my_function2, inputs=['inp2_1'], outputs=['inp3_2'])
    graph0.add_node(graph0_1, kwargs=[('inp1', 'inp1_1'), ('inp2', 'inp3_2')], outputs=['out'])

    graph0_0.add_node(f_my_function1, inputs=['inp1_1', 'inp1_2'], outputs=['inp2_1'])
    graph0_0.add_node(f_my_function2, inputs=['inp2_1'], outputs=['inp3_2'])
    graph0_0.add_node(f_my_function3, inputs=[('inp3_1', 'inp1_1'), 'inp3_2'], outputs=['out'])

    # inp2_1 = 0.215, inp3_2 = 0.0215, out = 0.1785
    graph0_1_0.add_node(f_my_function1, inputs=['inp1_1', 'inp1_2'], outputs=['inp2_1'])
    graph0_1_0.add_node(f_my_function2, inputs=['inp2_1'], outputs=['inp3_2'])
    graph0_1_0.add_node(f_my_function3, inputs=[('inp3_1', 'inp1_1'), 'inp3_2'], outputs=['out'])

    # out1 = 0.2, out2 = 0.015, out=0.1785
    graph0_1.add_node(f_my_function2, inputs=['inp1'], outputs=['out1'])
    graph0_1.add_node(f_my_function2, inputs=['inp2'], outputs=['out2'])
    graph0_1.add_node(graph0_1_0, kwargs=[('inp1_1', 'out1'), ('inp1_2', 'out2')], outputs=['out3'])

    prefix_ids, deep_ordered_nodes = zip(*graph0.deep_prefix_id_ordered_nodes())

    graph0_nodes = graph0.ordered_nodes
    graph0_0_nodes = graph0_0.ordered_nodes
    graph0_1_nodes = graph0_1.ordered_nodes
    graph0_1_0_nodes = graph0_1_0.ordered_nodes

    graph0_prefix_ids = [n.id for n in graph0.ordered_nodes]
    graph0_0_prefix_ids = [graph0_nodes[0].id + '.' + n.id for n in graph0_0.ordered_nodes]
    graph0_1_prefix_ids = [graph0_nodes[2].id + '.' + n.id for n in graph0_1.ordered_nodes]
    graph0_1_0_prefix_ids = [graph0_nodes[2].id + '.' + graph0_1_nodes[2].id + '.' + n.id
                             for n in graph0_1_0.ordered_nodes]

    assert deep_ordered_nodes[0] is graph0_nodes[0]
    for n1, n2 in zip(deep_ordered_nodes[1:4], graph0_0_nodes):
        assert n1 is n2
    assert deep_ordered_nodes[4] is graph0_nodes[1]
    assert deep_ordered_nodes[5] is graph0_nodes[2]
    for n1, n2 in zip(deep_ordered_nodes[6:9], graph0_1_nodes):
        assert n1 is n2
    for n1, n2 in zip(deep_ordered_nodes[9:], graph0_1_0_nodes):
        assert n1 is n2

    assert prefix_ids[0] == graph0_prefix_ids[0]
    for pi1, pi2 in zip(prefix_ids[1:4], graph0_0_prefix_ids):
        assert pi1 == pi2
    assert prefix_ids[4] == graph0_prefix_ids[1]
    assert prefix_ids[5] == graph0_prefix_ids[2]
    for pi1, pi2 in zip(prefix_ids[6:9], graph0_1_prefix_ids):
        assert pi1 == pi2
    for pi1, pi2 in zip(prefix_ids[9:], graph0_1_0_prefix_ids):
        assert pi1 == pi2



""" Main module containing Graph / Node classes """

from functools import reduce
import logging
import inspect
import copy
from multiprocessing.pool import Pool

from typing import Optional, Union, List, Callable, Dict, Tuple, Any

from alchemy_cat.dag.io import Input, Output
from alchemy_cat.dag.errors import PyungoError
from alchemy_cat.dag.utils import get_function_return_names, run_node
from alchemy_cat.dag.data import Data
from alchemy_cat.py_tools import Timer

logging.basicConfig()
LOGGER = logging.getLogger()
LOGGER.setLevel(logging.INFO)


def topological_sort(data):
    """ Topological sort algorithm

    Args:
        data (dict): dictionary representing dependencies
            Example: {'a': ['b', 'c']} node id 'a' depends on
            node id 'b' and 'c'

    Returns:
        ordered (list): list of list of node ids
            Example: [['a'], ['b', 'c'], ['d']]
            The sequence is representing the order to be run.
            The nested lists are node ids that can be run in parallel

    Raises:
        PyungoError: In case a cyclic dependency exists
    """
    for key in data:
        data[key] = set(data[key])
    for k, v in data.items():
        v.discard(k)  # ignore self dependencies
    extra_items_in_deps = reduce(set.union, data.values()) - set(data.keys())  # Add items without dependencies to data
    data.update({item: set() for item in extra_items_in_deps})
    while True:
        ordered = set(item for item, dep in data.items() if not dep)
        if not ordered:
            break
        yield sorted(ordered)
        data = {item: (dep - ordered) for item, dep in data.items()
                if item not in ordered}
    if data:
        raise PyungoError('A cyclic dependency exists amongst {}'.format(data))


class Node:
    """ Node object (aka vertex in graph theory)

    Args:
        fct (Callable): The Python function attached to the node
        outputs (list): List of outputs (`Output` or `str`)
        args (list): Optional list of args
        kwargs (list): Optional list of kwargs

    Raises:
        PyungoError: In case inputs have the wrong type
    """

    def __init__(self, fct: Callable, outputs=None, args=None, kwargs=None, verbose=False, slim_names=None):
        self._fct = fct
        self._verbose = verbose
        self._slim_names = slim_names if slim_names else []

        self._inputs = []
        self._outputs = []
        self._fct_defaults = {}

        self._args = args if args else []
        self._kwargs = kwargs if kwargs else []

        # if inputs are None, we inspect the function signature
        if (not self._args) and (not self._kwargs):
            self._args = list(inspect.signature(self._fct).parameters.keys())
            if not self._args:
                raise PyungoError(f"Node's func {self._fct} must have input params")

        self._output_names = outputs if outputs else []

        # if inputs are None, we inspect the function code
        if not self._output_names:
            self._output_names = get_function_return_names(self._fct)

        self._process_inputs(self._args)
        self._process_inputs(self._kwargs, is_kwarg=True)

        self._process_fct_defaults()

        self._process_outputs(self._output_names)

        self._id = str(self)

    def __repr__(self):
        return 'Node(<{}>, {}, {})'.format(self.fct_name, self.input_names, self.output_names)

    def __call__(self, *args, **kwargs):
        """ run the function attached to the node, and store the result """
        if self._verbose:
            timer = Timer().start()
        res = self._fct(*args, **kwargs)
        if self._verbose:
            timer.close()
            LOGGER.info('Ran {} in {}'.format(self, timer))
        return res

    @property
    def id(self):
        """ return the unique id of the node """
        return self._id

    # @property
    # def inputs(self):
    #     " return inputs"
    #     return self._inputs

    @property
    def input_names(self):
        """ return a list of all input names """
        input_names = [i.name for i in self._inputs]
        return input_names

    # @property
    # def input_maps(self):
    #     """ return a list of all input maps """
    #     input_maps = [i.map for i in self._inputs]
    #     return input_maps

    @property
    def inputs_without_constants(self):
        """ return the list of inputs, when inputs are not constants """
        inputs = [i for i in self._inputs if not i.is_constant]
        return inputs

    # @property
    # def kwargs_name(self):
    #     """ return the list of kwargs """
    #     return [i.name for i in self._inputs if i.is_kwarg]

    # @property
    # def outputs(self):
    #     return self._outputs

    @property
    def output_names(self):
        """ return a list of output names """
        return [o.name for o in self._outputs]

    @property
    def nec_input_maps(self):
        """ return necessary input names """
        return [i.map for i in self._inputs if (not i.is_constant) and (i.name not in self._fct_defaults)]

    @property
    def opt_input_maps(self):
        """ return optional input names """
        return [i.map for i in self._inputs if (not i.is_constant) and (i.name in self._fct_defaults)]

    @property
    def fct_name(self):
        """ return the function name """
        if hasattr(self._fct, '__name__'):
            fct_name = self._fct.__name__
        else:
            fct_name = type(self._fct).__name__  # For fct is functor
        return fct_name

    def _process_inputs(self, inputs, is_kwarg=False):
        """ converter data passed to Input objects and store them """
        if isinstance(inputs, dict):
            inputs = [{k: v} for k, v in inputs.items()]

        for inp in inputs:
            if isinstance(inp, tuple):
                if len(inp) != 2:
                    raise PyungoError(f"Tuple input should like (name, map). However, get {inp}")
                new_input = Input(name=inp[0], map=inp[1], is_kwarg=is_kwarg)
            elif isinstance(inp, Input):
                new_input = inp
                new_input.is_kwarg = is_kwarg
            elif isinstance(inp, str):
                new_input = Input(name=inp, is_kwarg=is_kwarg)
            elif isinstance(inp, dict):
                if len(inp) != 1:
                    msg = ('dict inputs should have only one key '
                           'and cannot be empty')
                    raise PyungoError(msg)
                key = next(iter(inp))
                value = inp[key]
                new_input = Input(name=key, value=value, is_kwarg=is_kwarg)
            else:
                msg = 'inputs need to be of type tuple, Input, str or dict'
                raise PyungoError(msg)
            self._inputs.append(new_input)

    def _process_fct_defaults(self):
        """ read and store kwargs default values"""
        self._fct_defaults = {k: v.default for k, v in inspect.signature(self._fct).parameters.items()
                              if v.default is not inspect.Parameter.empty}

    def _process_outputs(self, outputs):
        """ converter data passed to Output objects and store them """
        for output in outputs:
            if isinstance(output, Output):
                new_output = output
            elif isinstance(output, str):
                new_output = Output(output)
            else:
                msg = 'outputs need to be of type Outputs or str'
                raise PyungoError(msg)
            self._outputs.append(new_output)

    def run_with_loaded_inputs(self):
        """ Run the node with the attached function and loaded input values """
        args = [i.value for i in self._inputs if not i.is_kwarg]
        kwargs = {i.name: i.value for i in self._inputs if i.is_kwarg}
        rslts = self(*args, **kwargs)
        # Save outputs
        if len(self._outputs) == 1:
            self._outputs[0].value = rslts
        else:
            for output, rslt in zip(self._outputs, rslts):
                output.value = rslt
        return rslts


class Graph:
    """ Graph object, collection of related nodes

    Args:
        pool_size (int): Size of the pool. 0 means don't use parallel.
        schema (dict): Optional JSON schema to validate inputs data
        verbosity (int): 0: No log output; 1: Only give graph level log; >1: Give graph and node level log.
        slim (bool): If True, use copy rather than deepcopy when setting value of Node's input.

    Raises:
        ImportError will raise in case parallelism is chosen and `multiprocess`
            not installed
    """

    def __init__(self, pool_size: int = 0, schema: dict = None, verbosity: int = 0, slim: bool = False):
        self._nodes = {}
        self._data = None
        self._schema = schema

        self.pool_size = pool_size
        self.verbosity = verbosity
        self.slim = slim

        self._dag = None

        self._dag_output_names = set()
        self._dag_nec_input_maps = set()
        self._dag_opt_input_maps = set()

    @property
    def data(self):
        """ return the data of the graph (inputs + outputs) """
        return self._data

    # @property
    # def sim_inputs(self):
    #     """ return input names (mapped) of every nodes need feeding from Data """
    #     inputs = []
    #     for node in self._nodes.values():
    #         inputs.extend([i.map for i in node.inputs_without_constants
    #                        if not i.is_kwarg])
    #     return inputs
    #
    # @property
    # def sim_kwargs(self):
    #     """ return kwarg names (mapped) of every nodes """
    #     kwargs = [k for node in self._nodes.values() for k in node.kwargs_name]
    #     return kwargs
    #
    # @property
    # def sim_outputs(self):
    #     """ return output names (mapped) of every nodes """
    #     outputs = []
    #     for node in self._nodes.values():
    #         outputs.extend([o.map for o in node.outputs])
    #     return outputs

    def _dependencies(self):
        """ return dependencies among the nodes """
        dep = {}
        for node in self._nodes.values():
            d = dep.setdefault(node.id, [])
            for inp in node.inputs_without_constants:
                for node2 in self._nodes.values():
                    if inp.map in node2.output_names:
                        d.append(node2.id)
        return dep

    @property
    def dependencies(self):
        """Return dependencies of each node"""
        return self._dependencies()

    @property
    def dag(self):
        """ If dag is not built, build dag. Return the ordered nodes graph """
        if self._dag is None:
            dag = []
            for node_ids in topological_sort(self._dependencies()):
                nodes = [self.get_node(node_id) for node_id in node_ids]
                dag.append(nodes)
            self._dag = dag
        return self._dag

    @property
    def ordered_nodes(self) -> List[Node]:
        """Same to the dag except returned nodes is 1-Dimension"""
        ordered_nodes = []
        for nodes in self.dag:
            ordered_nodes.extend(nodes)
        return ordered_nodes

    @property
    def deep_ordered_nodes(self):
        """Return all nodes in graph (include sub-graph's nodes) according to their running start order"""
        deep_ordered_nodes = []
        for node in self.ordered_nodes:
            deep_ordered_nodes.append(node)
            if isinstance(node._fct, Graph):
                deep_ordered_nodes.extend(node._fct.deep_ordered_nodes)
        return deep_ordered_nodes

    def deep_prefix_id_ordered_nodes(self, prefix: str= ''):
        """Return all (prefix_id, nodes) in graph (include sub-graph's nodes) according to their running start order"""
        ret = []
        for node in self.ordered_nodes:
            prefix_id = prefix + '.' + node.id if prefix else node.id
            ret.append((prefix_id, node))
            if isinstance(node._fct, Graph):
                ret.extend(node._fct.deep_prefix_id_ordered_nodes(prefix=prefix_id))
        return ret

    @staticmethod
    def run_node(node):
        """ run the node

        Args:
            node (Node): The node to run

        Returns:
            results (tuple): node id, node output values

        Raises:
            PyungoError
        """
        return node.id, node.run_with_loaded_inputs()

    def _register(self,
                  f: Union[Callable, type],
                  inputs: Optional[
                      Union[List[Union[str, Tuple[str, str], Input, Dict[str, Any]]], Dict[str, Any]]] = None,
                  outputs: Optional[List[Union[str, Output]]] = None,
                  args: Optional[
                      Union[List[Union[str, Tuple[str, str], Input, Dict[str, Any]]], Dict[str, Any]]] = None,
                  kwargs: Optional[
                      Union[List[Union[str, Tuple[str, str], Input, Dict[str, Any]]], Dict[str, Any]]] = None,
                  slim_names: Optional[List[str]] = None,
                  init: Optional[Dict[str, Any]] = None):
        """ get provided inputs, outputs, args, kwargs, slim_names, init if any and create a new node """
        if not callable(f):
            raise PyungoError(f"Registered function {f} should be callable")

        if (inputs is not None) and (args is not None):
            args = inputs + args
        elif inputs is not None:
            args = inputs

        # Instantiate f if f is functor
        if inspect.isclass(f):
            init = init if init is not None else {}
            f = f(**init)
        elif init is not None:
            raise PyungoError("Only functor can be initialized with 'init'")

        if isinstance(f, Graph):
            if args is not None:
                raise PyungoError(f"Node with Graph can only accept kwargs input. However, get args = {args}")

        # create and save the node to the graph
        node = Node(f, outputs, args, kwargs, True if self.verbosity > 1 else False, slim_names)
        # assume that we cannot have two nodes with the same output names
        diff = set(node.output_names) & self._dag_output_names
        if diff:
            raise PyungoError(f"Node {node} have repeated output names: {sorted(list(diff))}")

        # assume no self dependencies
        diff = set([inp.map for inp in node.inputs_without_constants]) & set(node.output_names)
        if diff:
            raise PyungoError(f"Node {node} have self dependence caused by the following inputs: {sorted(list(diff))}")

        self._dag_output_names |= set(node.output_names)
        self._dag_nec_input_maps |= set(node.nec_input_maps)
        self._dag_opt_input_maps |= set(node.opt_input_maps)
        self._dag_opt_input_maps -= self._dag_nec_input_maps

        self._nodes[node.id] = node
        self._dag = None

    def register(self,
                 inputs: Optional[
                     Union[List[Union[str, Tuple[str, str], Input, Dict[str, Any]]], Dict[str, Any]]] = None,
                 outputs: Optional[List[Union[str, Output]]] = None,
                 args: Optional[Union[List[Union[str, Tuple[str, str], Input, Dict[str, Any]]], Dict[str, Any]]] = None,
                 kwargs: Optional[
                     Union[List[Union[str, Tuple[str, str], Input, Dict[str, Any]]], Dict[str, Any]]] = None,
                 slim_names: Optional[List[str]] = None,
                 init: Optional[Dict[str, Any]] = None):
        """ register decorator

            Args:
                function : Python function attached to the node
                inputs (list): List of inputs (Input, str, or dict)
                outputs (list): List of outputs (Output or str)
                args (list): List of optional args
                kwargs (list): List of optional kwargs
                slim_names (list): List of args which should use copy rather than deepcopy
                init (dict): init dict for functor
        """

        def decorator(f):
            self._register(f, inputs, outputs, args, kwargs, slim_names, init)
            return f

        return decorator

    def add_node(self,
                 function: Union[Callable, type],
                 inputs: Optional[
                     Union[List[Union[str, Tuple[str, str], Input, Dict[str, Any]]], Dict[str, Any]]] = None,
                 outputs: Optional[List[Union[str, Output]]] = None,
                 args: Optional[Union[List[Union[str, Tuple[str, str], Input, Dict[str, Any]]], Dict[str, Any]]] = None,
                 kwargs: Optional[
                     Union[List[Union[str, Tuple[str, str], Input, Dict[str, Any]]], Dict[str, Any]]] = None,
                 slim_names: Optional[List[str]] = None,
                 init: Optional[Dict[str, Any]] = None):
        """ explicit method to add a node to the graph

        Args:
            function : Python function attached to the node
            inputs (list): List of inputs (Input, str, or dict)
            outputs (list): List of outputs (Output or str)
            args (list): List of optional args
            kwargs (list): List of optional kwargs
            slim_names (list): List of args which should use copy rather than deepcopy
            init (dict): init dict for functor
        """
        self._register(function, inputs, outputs, args, kwargs, slim_names, init)

    def get_node(self, id_: str) -> Node:
        """ get a node from its id """
        return self._nodes[id_]

    def deep_get_node(self, prefix_id: str):
        """ get a node by it's prefix_id"""
        graph = self
        ids = prefix_id.split('.')
        for id in ids[:-1]:
            graph = graph.get_node(id)._fct
            if not isinstance(graph, Graph):
                raise PyungoError(f"For node id = {id} in prefix_id = {prefix_id}, "
                                  f"it's fct has type {type(graph)} rather than Graph")
        return graph.get_node(ids[-1])

    def calculate(self, data):
        """ run graph calculations """
        # make sure data is valid when using schema
        if self._schema:
            try:
                import jsonschema
            except ImportError:
                msg = 'jsonschema package is needed for validating data'
                raise ImportError(msg)
            jsonschema.validate(instance=data, schema=self._schema)

        if self.verbosity:
            timer = Timer().start()
            LOGGER.info('Starting calculation...')

        self._data = Data(data, self.slim)
        self._data.check_inputs(self._dag_nec_input_maps, self._dag_opt_input_maps, self._dag_output_names)

        def set_node_input_value(node, input, value):
            if (not self.slim) and (input.name not in node._slim_names):
                input.value = copy.deepcopy(value)
            else:
                input.value = value

        for nodes in self.dag:
            # loading node with inputs
            for node in nodes:
                for inp in node.inputs_without_constants:
                    if inp.map in self._data:
                        set_node_input_value(node, inp, self._data[inp.map])
                    else:
                        set_node_input_value(node, inp, node._fct_defaults[inp.name])

            # running nodes
            if self.pool_size:
                pool = Pool(self.pool_size)
                rslts = pool.map(run_node, nodes)  # ! The output of node's won't be copied back
                pool.close()
                pool.join()

                for node, rslt in zip(nodes, rslts):
                    if len(node._outputs) == 1:
                        node._outputs[0].value = rslt
                    else:
                        for output, r in zip(node._outputs, rslt):
                            output.value = r
            else:
                for node in nodes:
                    node.run_with_loaded_inputs()

            # save results
            for node in nodes:
                for output in node._outputs:
                    self._data[output.map] = output.value

        if self.verbosity:
            timer.close()
            LOGGER.info(f'Calculation finished in {timer}')

        last_node_outputs = self.dag[-1][-1]._outputs
        if len(last_node_outputs) == 1:
            return last_node_outputs[0].value
        else:
            return tuple(output.value for output in last_node_outputs)

    def __call__(self, *args, **kwargs):
        "Equal to graph.calculate(kwargs)"
        if args:
            raise PyungoError("Graph only receive keyword args which will be recognized as input name and value.")

        return self.calculate(kwargs)

    def __repr__(self):
        return f"Graph with {len(self._nodes)} nodes in " + \
               ("slim mode" if self.slim is True else "non-slim mode") + \
               '.'

    def __str__(self):
        ret = repr(self) + "\n"
        for prefix_id, node in self.deep_prefix_id_ordered_nodes():
            ret += f"{prefix_id}\n"
            if isinstance(node._fct, Graph):
                ret += "--> " + repr(node._fct) + "\n"
        return ret



import numpy as np

from alchemy_cat.dag.core import Graph

graph = Graph(pool_size=2, verbosity=2, slim=False)

@graph.register()
def f_my_function(a, b):
    c = a + b
    return c
@graph.register(inputs=['d', 'a'], outputs=['e'])
def f_my_function3(d, a):
    e = d - a
    return e
@graph.register(outputs=['d'])
def f_my_function2(c):
    return c / 10.

if __name__ == '__main__':
    res = graph.calculate(data={'a': np.random.randn(2000, 2000), 'b': np.random.randn(2000, 2000)})
    print(res)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/5/24 16:25
@File    : seg_mask_metric.py
@Software: PyCharm
@Desc    : 
"""
import numpy as np

from alchemy_cat.contrib.metrics.conf_matrix_bases_metrics import SegmentationMetric
from alchemy_cat.py_tools import Statistic

__all__ = ["SegMaskMetric"]



class SegMaskMetric(SegmentationMetric):
    """SegMask Metric which implement ignore label in pred as a independent class."""

    @Statistic.getter(1)
    def macro_avg_precision(self):
        return np.nanmean(self.cls_precision[:-1])

    @Statistic.getter(1)
    def fg_precision(self):
        return np.nanmean(self.cls_precision[1:-1])

    @Statistic.getter(1)
    def fg_recall(self):
        return np.nanmean(self.cls_recall[1:])

    @Statistic.getter(1)
    def fg_mIoU(self):
        return np.mean(self.cls_IoU[1:][self.valid[1:]])

    @Statistic.getter(1)
    def bg_precision(self):
        return self.cls_precision[0]

    @Statistic.getter(1)
    def bg_recall(self):
        return self.cls_recall[0]

    @Statistic.getter(1)
    def bg_mIoU(self):
        return self.cls_IoU[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/3/13 16:37
@File    : seed_thresh.py
@Software: PyCharm
@Desc    : 
"""
from typing import Callable

import numpy as np
import torch

from .score import cam2score, cam2score_cuda, cat_bg_score, cat_bg_score_cuda, idx2seed, idx2seed_cuda

__all__ = ['seed_argmax', 'seed_argmax_cuda']


def _parse_dsize(dsize):
    if isinstance(dsize, np.ndarray) and dsize.ndim == 3:  # dsize是HWC的图片。
        return dsize.shape[:2], dsize
    else:
        return dsize, None


def seed_argmax(cam: np.ndarray, dsize, fg_cls: np.ndarray, bg_method: dict, resize_first: bool,
                crf: Callable[[np.ndarray, np.ndarray], np.ndarray] | None=None) -> (np.ndarray, np.ndarray):
    dsize, img = _parse_dsize(dsize)

    score = cam2score(cam, dsize=dsize, resize_first=resize_first)

    score = cat_bg_score(score, bg_method)

    if crf is not None:
        score = crf(img, score)

    seed = np.argmax(score, axis=0)
    seed = idx2seed(seed, fg_cls).astype(np.uint8)

    return seed, score


# TODO 支持batch输入。
def seed_argmax_cuda(cam: np.ndarray | torch.Tensor, dsize, fg_cls: np.ndarray | torch.Tensor,
                     bg_method: dict, resize_first: bool,
                     crf: Callable[[np.ndarray, np.ndarray], np.ndarray] | None=None) -> (np.ndarray, np.ndarray):
    dsize, img = _parse_dsize(dsize)

    cam = torch.asarray(cam, device='cuda', requires_grad=False)
    fg_cls = torch.asarray(fg_cls, device='cuda', requires_grad=False)

    score = cam2score_cuda(cam, dsize=dsize, resize_first=resize_first)

    score = cat_bg_score_cuda(score, bg_method)

    if crf is not None:
        score = torch.from_numpy(crf(img, score.cpu().numpy())).to(device='cuda')

    seed = torch.argmax(score, dim=0)
    seed = idx2seed_cuda(seed, fg_cls).to(torch.uint8)

    return seed.cpu().numpy(), score.cpu().numpy()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/8/2 21:09
@File    : py3_server.py
@Software: PyCharm
@Desc    : 
"""
import numpy as np

from alchemy_cat.contrib.shm_send_recv.py3_send_recv import server_send, server_recv, init_server  # 导入server所需函数

kImageSize = (512, 1024)
kServerPredict = np.random.randint(0, 256, size=kImageSize, dtype=np.uint8)


init_server()  # 初始化server，非必须。可以防止多个server同时运行，且保证server被中断后（譬如被kill）后，buffer能被正确清理。

while True:
    img = server_recv()  # 接收图片

    predict = kServerPredict  # 根据输入图片，得到预测结果。这里用常量代替。

    server_send(kServerPredict)  # 发送预测结果



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/11/24 22:51
@File    : file_id_dt.py
@Software: PyCharm
@Desc    : 
"""
import os.path as osp

import cv2

from alchemy_cat.data import Dataset
from alchemy_cat.py_tools.find_files import find_files_by_exts

__all__ = ["FileIDDt", "ImgIDDt"]


class FileIDDt(Dataset):
    """Dataset organized with file id"""

    def __init__(self, root_dir: str, file_ext: str):
        """
        Args:
            root_dir (str): root dir.
            file_ext (str): File extension map, can be '.png', '.jpg', etc.
        """
        self.root_dir = root_dir
        self.file_ext = file_ext

        self.files = None
        self.file_ids = None
        self._set_files()

    def _set_files(self):
        self.files = list(find_files_by_exts(self.root_dir, [self.file_ext]))
        self.files.sort()
        self.file_ids = [osp.basename(img_file).split('.')[0] for img_file in self.files]

    def __len__(self):
        return len(self.file_ids)

    def get_item(self, index):
        return self.file_ids[index], self.load_file(self.files[index])

    def load_file(self, file):
        raise NotImplementedError

    def get_by_id(self, file_id: str):
        try:
            index = self.file_ids.index(file_id)
        except ValueError:
            raise RuntimeError(f"Can't find file_id {file_id} in dataset's file_ids list.")
        return self[index]


class ImgIDDt(FileIDDt):
    """Dataset organized with img_id"""

    def __init__(self, root_dir: str, file_ext: str = '.png', cv2_read_flag: int = cv2.IMREAD_GRAYSCALE):
        super().__init__(root_dir, file_ext)
        self.cv2_read_flag = cv2_read_flag

    def load_file(self, file):
        img = cv2.imread(file, self.cv2_read_flag)
        assert img is not None
        return img



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: val_env.py
@time: 2020/3/29 3:17
@desc:
"""
import cv2
import torch
import torch.nn as nn
from alchemy_cat.py_tools.random import set_torch_rand_seed

from .rng_cacher import RNGCacher

__all__ = ['ValidationEnv']

class ValidationEnv(object):
    """Context manager for validate model"""
    def __init__(self, model: nn.Module,
                 change_benchmark: bool=False, seed: int=0, empty_cache: bool=True, cv2_num_threads: int=-1):
        """Context manager for validate model

        Args:
            model: model to be validated
            change_benchmark: If True, will inverse torch.backends.cudnn.benchmark before and after context.
                (Default: False)
            seed: Random seed for validation. (Default: 0)
            empty_cache: If True, will empty cuda cache before validation. (Default: True)
            cv2_num_threads: Number of threads used by opencv.
        """
        self.model = model

        self.change_benchmark = change_benchmark
        self.seed = seed
        self.empty_cache = empty_cache
        self.cv2_num_threads = cv2_num_threads

        self.out_grad_enabled: bool | None = None
        self.cacher = RNGCacher()
        self.out_model_training: bool | None = None
        self.out_cv2_num_threads: int | None = None

    def __enter__(self) -> nn.Module:
        # -* 尽量释放显存给推理。
        if self.empty_cache:
            torch.cuda.empty_cache()
        # -* 记录外部梯度可用性，关闭梯度。
        self.out_grad_enabled = torch.is_grad_enabled()
        torch.set_grad_enabled(False)
        # -* 记录外部随机性。
        self.cacher.cache()
        # -* 重设随机种子。
        set_torch_rand_seed(self.seed)
        # -* 记录外部模型是否可训，并设置为验证状态。
        self.out_model_training = self.model.training
        self.model.eval()
        # -* 改变benchmark。
        if self.change_benchmark:
            torch.backends.cudnn.benchmark = not torch.backends.cudnn.benchmark
        # -* 记录cv2线程数。
        self.out_cv2_num_threads = cv2.getNumThreads()
        cv2.setNumThreads(self.cv2_num_threads)
        return self.model

    def __exit__(self, exc_type, exc_val, exc_tb):
        # -* 恢复cv2线程数。
        cv2.setNumThreads(self.out_cv2_num_threads)
        # * 恢复benchmark。
        if self.change_benchmark:
            torch.backends.cudnn.benchmark = not torch.backends.cudnn.benchmark
        # * 恢复外部模型可训性。
        self.model.train(self.out_model_training)
        # * 恢复外部随机状态。
        self.cacher.resume(purge=True)
        # * 恢复外部梯度可用性。
        torch.set_grad_enabled(self.out_grad_enabled)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/5/24 20:05
@File    : open_cfg.py
@Software: PyCharm
@Desc    : 
"""
import glob
import json
import pickle
from os import path as osp
from pathlib import Path
from typing import Tuple, Union

try:
    from yacs.config import CfgNode, _assert_with_logging, _valid_type, _VALID_TYPES  # noqa
    YACS_AVAILABLE = True
except ImportError:
    YACS_AVAILABLE = False

try:
    from mmengine import Config as MMConfig  # noqa
    MM_AVAILABLE = True
except ImportError:
    MM_AVAILABLE = False

from ..load_module import load_module_from_py

__all__ = ['open_config']


def open_config(config_path: Union[str, dict], is_yaml: bool = False) -> Tuple[dict, bool]:
    """根据文件路径载入配置树所在模块，并读取名为config的配置树。

    Args:
        config_path (str): 配置树模块路径。
        is_yaml (bool): 过时的无效参数。

    Returns: 配置树模块中名为config的配置树。
    """
    # * is_yaml参数已经过时。
    if is_yaml:
        raise DeprecationWarning("is_yaml parameter is deprecated. ")

    # * 若config_path是dict，则直接返回。
    if isinstance(config_path, dict):
        return config_path, True

    # * Get extension
    name, ext = osp.splitext(config_path)
    # ** auto ext
    if ext == '':
        cfg_path_found = glob.glob(f'{name}.*', recursive=False)
        assert len(cfg_path_found) == 1, f"Found {len(cfg_path_found)} config files with {name=}: {cfg_path_found}"
        config_path = cfg_path_found[0]
        _, ext = osp.splitext(config_path)

    # * Check for existence
    if (ext != '.py') and (not osp.isfile(config_path)):  # py文件找不到，可能是从其他路径导入。
        raise RuntimeError(f"No config file found at {config_path}")

    is_py = True
    # * Read yml/yaml config
    if ext == '.yml' or ext == '.yaml':
        # * Read and set config
        import yaml
        config = yaml.safe_load(Path(config_path).read_text())
    elif ext == '.py':
        cfg_module = load_module_from_py(config_path)
        if hasattr(cfg_module, 'config'):
            config: dict = cfg_module.config
            if hasattr(cfg_module, 'cfg') and (cfg_module.cfg is not cfg_module.config):
                raise RuntimeError(f"Both 'config' and 'cfg' are defined in {cfg_module.__path__} but not the same.")
        elif hasattr(cfg_module, 'cfg'):
            config: dict = cfg_module.cfg
        elif hasattr(cfg_module, 'get_cfg_defaults'):  # yacs兼容
            config: dict = cfg_module.get_cfg_defaults()
        elif hasattr(cfg_module, '_C'):  # yacs兼容
            config: dict = getattr(cfg_module, '_C')
        elif MM_AVAILABLE:
            mm_config: MMConfig = MMConfig.fromfile(config_path)
            if hasattr(mm_config, 'to_dict'):
                config: dict = mm_config.to_dict()
            else:
                config: dict = object.__getattribute__(mm_config, '_cfg_dict').to_dict()
        else:
            raise ValueError(f"config module {cfg_module.__path__} must have 'config'/'cfg' defined, "
                             f"or be a yacs or mmcv config.")

        # -* 如果是yacs配置，则转为dict。
        if YACS_AVAILABLE and isinstance(config, CfgNode):
            def convert_to_dict(cfg_node, key_list):
                if not isinstance(cfg_node, CfgNode):
                    _assert_with_logging(
                        _valid_type(cfg_node),
                        "Key {} with value {} is not a valid type; valid types: {}".format(
                            ".".join(key_list), type(cfg_node), _VALID_TYPES
                        ),
                    )
                    return cfg_node
                else:
                    cfg_dict = dict(cfg_node)
                    for k, v in cfg_dict.items():
                        cfg_dict[k] = convert_to_dict(v, key_list + [k])
                    return cfg_dict

            config = convert_to_dict(config, [])
    elif ext == '.json':
        with open(config_path, 'r') as json_f:
            config = json.load(json_f)
    elif ext == ".pkl":
        with open(config_path, 'rb') as pkl_f:
            config: dict = pickle.load(pkl_f)
    else:
        raise ValueError(f"config_path = {config_path} must be python, json, pkl, yml or yaml file.")

    if not isinstance(config, dict):
        raise ValueError(f"The opened {config_path} should be a dict, but got {type(config)}")

    return config, is_py



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/6/18 17:36
@File    : utils.py
@Software: PyCharm
@Desc    : 
"""
from typing import Any, Union, Final
import os.path as osp

__all__ = ['norm_param_name', 'name_param_val']

kLongestParamStr: Final[int] = 75


def norm_param_name(name: str) -> str:
    return name.replace('.', '·').replace(osp.sep, '-')  # value中的小数点用·替代，以免影响导入；/用-替代，以免影响路径。


def _is_int(elem: Any) -> bool:
    """拷贝自 alchemy_cat/py_tools/type.py，支持numpy、torch未安装情况下运行。"""
    try:
        import torch
        TORCH_AVAILABLE = True
    except ImportError:
        torch = None
        TORCH_AVAILABLE = False

    try:
        import numpy as np
        NUMPY_AVAILABLE = True
    except ImportError:
        np = None
        NUMPY_AVAILABLE = False

    if TORCH_AVAILABLE and isinstance(elem, torch.Tensor):
        if elem.ndim != 0:
            return False
        elif (not torch.is_floating_point(elem)) and (not torch.is_complex(elem)) and (elem.dtype is not torch.bool):
            return True
        else:
            return False

    if NUMPY_AVAILABLE and isinstance(elem, np.integer):
        return True

    if isinstance(elem, int) and (not isinstance(elem, bool)):
        return True

    return False


def name_param_val(param_val: Any, longest_param_length: int=kLongestParamStr) -> Union[str, int, float]:
    # if is_int(param_val) or is_float(param_val):  # int或float保持不变，适配pandas索引。
    if _is_int(param_val):  # NOTE 只有int保持不变，适配pandas索引。float中的小数点会影响导入，还是需要替换。
        return param_val

    param_val_str = str(param_val).splitlines()[0]
    if isinstance(param_val, dict) and '_param_val_name' in param_val:
        ret = str(param_val['_param_val_name'])
    elif callable(param_val) and hasattr(param_val, '__name__') and len(param_val.__name__) <= longest_param_length:
        ret = param_val.__name__
    elif len(param_val_str) <= longest_param_length:
        ret = param_val_str
    else:
        ret = str(hash(param_val))[:longest_param_length]

    ret = norm_param_name(ret)

    return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/11/26 21:54
@File    : typing.py
@Software: PyCharm
@Desc    : 
"""
import typing as t

import numpy as np
import torch

__all__ = ['SegDataset']


T_ret = t.TypeVar('T_ret')


class SegDataset(t.Protocol[T_ret]):
    class_names: t.Sequence
    class_num: int
    ignore_label: int
    image_ids: t.Sequence
    split: str
    ret_img_file: bool

    def __getitem__(self, index: slice | list | np.ndarray | torch.Tensor | int) -> T_ret:
        ...

    def __add__(self, other):
        ...

    def get_item(self, index) -> T_ret:
        ...

    def get_by_img_id(self, img_id) -> T_ret:
        ...

    def __len__(self):
        ...

    def __repr__(self):
        ...

    def __iter__(self):
        ...

    @staticmethod
    def label_map2color_map(label_map: np.ndarray) -> np.ndarray:
        ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: find_files.py
@time: 2021/9/16 14:06
@desc: 
"""
from typing import List, Callable

import os
import os.path as osp
import re

__all__ = ['IMG_EXTS', 'find_files', 'find_files_by_exts', 'find_files_by_names', 'find_files_by_patterns',
           'find_img_files']


IMG_EXTS = ['.png', '.jpg', '.JPG', '.PNG', '.JPEG', '.jpeg']


def _match_exts(name_ext: str, exts: List[str]) -> bool:
    ext = osp.splitext(name_ext)[-1]
    return ext in exts


def _match_names(name_ext: str, names: List[str]) -> bool:
    name = osp.splitext(name_ext)[0]
    return name in names


def _match_patterns(name_ext: str, patterns: List[str]) -> bool:
    for pattern in patterns:
        if re.search(pattern, name_ext) is not None:
            return True
    return False


def find_files(root: str, is_match: Callable[[str], bool], recursion: bool=True, memo: str= '',
               is_full_path: bool=True) -> str:
    """Find files by function.

    Args:
        root: Root dir.
        is_match: Function receive files name_ext and return bool to indicate whether found.
        recursion: If True, find files recursively. (Default: True)
        memo: For recursion. (Default: '')
        is_full_path: If True return full path of files found. (Default: True)

    Returns:
        Yields find files under root.
    """
    name_exts = os.listdir(osp.join(root, memo))
    for name_ext in name_exts:
        full_path = osp.join(root, memo, name_ext)
        if osp.isdir(full_path) and recursion:
            yield from find_files(root, is_match, recursion, osp.join(memo, name_ext), is_full_path)
        elif osp.isfile(full_path) and is_match(name_ext):
            yield full_path if is_full_path else name_ext
        else:
            pass


def find_files_by_exts(root: str, exts: List[str], recursion: bool=True, memo: str= '', is_full_path: bool=True) -> str:
    """Find files by function.

    Args:
        root: Root dir.
        exts: Extensions wanted.
        recursion: If True, find files recursively. (Default: True)
        memo: For recursion. (Default: '')
        is_full_path: If True return full path of files found. (Default: True)

    Returns:
        Yields find files under root.
    """
    return find_files(root, lambda name_ext: _match_exts(name_ext, exts), recursion, memo, is_full_path)


def find_files_by_names(root: str, names: List[str], recursion: bool=True, memo: str= '',
                        is_full_path: bool=True) -> str:
    """Find files by function.

    Args:
        root: Root dir.
        names: Names wanted.
        recursion: If True, find files recursively. (Default: True)
        memo: For recursion. (Default: '')
        is_full_path: If True return full path of files found. (Default: True)

    Returns:
        Yields find files under root.
    """
    return find_files(root, lambda name_ext: _match_names(name_ext, names), recursion, memo, is_full_path)


def find_files_by_patterns(root: str, patterns: List[str], recursion: bool=True, memo: str= '', 
                           is_full_path: bool=True) -> str:
    """Find files by function.

    Args:
        root: Root dir.
        patterns: Patterns wanted.
        recursion: If True, find files recursively. (Default: True)
        memo: For recursion. (Default: '')
        is_full_path: If True return full path of files found. (Default: True)

    Returns:
        Yields find files under root.
    """
    return find_files(root, lambda name_ext: _match_patterns(name_ext, patterns), recursion, memo, is_full_path)


def find_img_files(root: str, recursion: bool=True, is_full_path: bool=True) -> str:
    """Find all image files.

    Args:
        root: Root dir.
        recursion: If True, find files recursively. (Default: True)
        is_full_path: If True return full path of files found. (Default: True)

    Returns:
        Yields find images under root.
    """
    return find_files_by_exts(root, IMG_EXTS, recursion, is_full_path=is_full_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/6/16 17:10
@File    : viz_seed.py
@Software: PyCharm
@Desc    : 
"""
import matplotlib.pyplot as plt
import numpy as np

from alchemy_cat.contrib.voc import label_map2color_map

__all__ = ['viz_seed']


def viz_seed(fig: plt.Figure, img_id: str, bottom_img: np.ndarray, label: np.ndarray, seed: np.ndarray,
             blend_alpha: float=0.7, color_map: callable=label_map2color_map):
    """ 可视化种子点。作图结果为：图像+label，图像+种子点，种子点。

    Args:
        fig: plt的Figure对象。
        img_id: 图像的编号。
        bottom_img: 图像数组。
        label: 标签数组。
        seed: 种子点数组。
        blend_alpha: 叠加图像时的透明度。
        color_map: 函数，输入单通道的数值标签，输出3通道的彩色标签。

    Returns:
        plt的Figure对象。
    """
    # * Show bottom_img + gt
    ax: plt.Axes = fig.add_subplot(3, 1, 1)
    ax.imshow(bottom_img)
    ax.imshow(color_map(label), alpha=blend_alpha)
    ax.set_title(img_id, fontsize='smaller')
    ax.axis("off")

    # * Show bottom_img + seg_mask
    ax: plt.Axes = fig.add_subplot(3, 1, 2)
    ax.imshow(bottom_img)
    ax.imshow(color_map(seed), alpha=blend_alpha)
    ax.axis("off")

    # * Show seg_mask only
    ax: plt.Axes = fig.add_subplot(3, 1, 3)
    ax.imshow(color_map(seed))
    ax.axis("off")

    fig.tight_layout(pad=0.2, h_pad=0.2, w_pad=0.2)
    return fig



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/7/7 14:56
@File    : cuda_block.py
@Software: PyCharm
@Desc    : 阻塞直至指定CUDA设备释放。
"""
import os
from typing import Union

from time import sleep

from gpustat import new_query, GPUStat

from alchemy_cat.py_tools import yprint, gprint

__all__ = ['parse_cudas', 'cudas2CUDA_VISIBLE_DEVICES', 'block_get_available_cuda', 'get_cudas']

kSleepSecs = 30


def parse_cudas(cudas: Union[list[Union[str, int]], str]) -> list[int]:
    if isinstance(cudas, str):
        cudas = [int(cuda) for cuda in cudas.split(',')]
    else:
        cudas = [int(cuda) for cuda in cudas]
    return cudas


def cudas2CUDA_VISIBLE_DEVICES(cudas: Union[list[Union[str, int]], str]) -> str:
    return ','.join([str(cuda) for cuda in parse_cudas(cudas)])


def block_get_available_cuda(cudas: Union[list[Union[str, int]], str],
                             cuda_need: int=-1, memory_need: float=-1., max_process: int=-1,
                             sleep_secs: int=kSleepSecs, verbosity: bool=True,
                             cudas_prefer: list[int]=None) -> list[int]:
    """Block the program util find enough available CUDA devices.

    Args:
        cudas: The list of CUDA devices can be represented as a list of strings (e.g., ['1', '2']),
            a list of integers (e.g., [0, 3]), or a string (e.g., '0,1,2').
        cuda_need: GPU need for available GPU. < 0 means need all GPU.
        memory_need: Memory need for available GPU (MB). < -0.5 means need 95% GPU memory.
        max_process: Max process num on available GPU. < 0 means no limit.
        sleep_secs: If block, sleep sleep_secs seconds and recheck.
        verbosity: Where print heartbeat.
        cudas_prefer: Prefer to use these cudas.

    Returns:
        Available CUDA devices.
    """
    # -* 获取常量。
    memory_total = new_query()[0].memory_total

    # -* 参数预处理。
    # -** cudas: list[int]
    cudas = parse_cudas(cudas)

    # -** 检查、解算gpu_need。此后 1 <= cuda_need <= len(cudas)。
    if cuda_need == 0:
        return []
    elif cuda_need > 0:
        assert len(cudas) >= cuda_need, f"Need {cuda_need} GPU, but only {len(cudas)} GPU available."
    else:
        cuda_need = len(cudas)

    # -** 检查、解算memory_need。此后 0 <= memory_need <= memory_total。
    assert memory_need <= memory_total, f"Memory need {memory_need}MB is larger than total memory {memory_total}MB."
    if memory_need < -0.5:  # 需要95%的显存。
        memory_need = memory_total * 0.95

    # -** 检查、解算max_process。此后 0 <= max_process。
    if max_process < 0:
        max_process = float('inf')

    # -* 遍历，直到有足够的空闲GPU。
    while True:
        gpu_states: list[GPUStat] = list(new_query())
        available_cudas: list[int] = []

        for c, gpu_state in ((cuda, gpu_states[cuda]) for cuda in cudas):
            if (gpu_state.memory_available >= memory_need) and (len(gpu_state.processes) <= max_process):
                available_cudas.append(c)

        if len(available_cudas) >= cuda_need:
            break
        else:
            yprint(f"[BLOCKING] {len(available_cudas)} cudas {available_cudas} in {cudas} available, "
                   f"waiting {sleep_secs}s. \n"
                   f"GPU need: {cuda_need}, Memory need: {memory_need:.1f}MB, Max process: {max_process}.")
            sleep(sleep_secs)

    if cudas_prefer is None:
        ret = available_cudas[:cuda_need]
    else:
        ret = set(cudas_prefer) & set(available_cudas)  # 找到prefer中可用的cuda。
        if len(ret) < cuda_need:  # 若可用cuda不足。
            ret = list(ret) + list(set(available_cudas) - ret)[:cuda_need-len(ret)]
        else:  # 若可用cuda足够。
            ret = list(ret)[:cuda_need]
    if verbosity:
        gprint(f"[SUCCESS] Get available cudas {ret}.")
    return ret


def get_cudas() -> tuple[list[int], str]:
    """Get CUDA_VISIBLE_DEVICES environment variable. If not exist, return all device.

    Returns:
        tuple[list of cuda index, CUDA_VISIBLE_DEVICES environment variable]
    """
    if 'CUDA_VISIBLE_DEVICES' in os.environ:
        cuda_env = os.environ['CUDA_VISIBLE_DEVICES']
        return parse_cudas(cuda_env), cuda_env
    else:
        cudas = [c.index for c in new_query()]
        return cudas, cudas2CUDA_VISIBLE_DEVICES(cudas)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2021/11/25 22:50
@File    : __init__.py.py
@Software: PyCharm
@Desc    : 
"""
from alchemy_cat.contrib.tasks.wsss.seeding.e2e_seeding.cam_sal_to_seed import *
from alchemy_cat.contrib.tasks.wsss.seeding.e2e_seeding.online_seed import *
from alchemy_cat.contrib.tasks.wsss.seeding.e2e_seeding.resolve_loc_cue_conflict import *
from alchemy_cat.contrib.tasks.wsss.seeding.e2e_seeding.one_hot_numeric_seed_convert import *
from alchemy_cat.contrib.tasks.wsss.seeding.e2e_seeding.label2cls_in_label import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: str_formatters.py
@time: 2020/2/6 11:43
@desc:
"""
from typing import Any
import time


__all__ = ['indent', 'get_local_time_str']


def indent(lines: Any, spaces: int=2) -> str:
    """Reformat input lines with spaces indent before each line

    Args:
        lines: lines to be added indents before. If not str, lines = str(lines)
        spaces: spaces in each indent. Default 2.

    Returns:
        Lines with indent
    """
    if not isinstance(lines, str):
       lines = str(lines)

    lines = lines.split('\n')
    indent = ' ' * spaces
    lines = [indent + line for line in lines]

    return '\n'.join(lines)


def get_local_time_str(for_file_name=False) -> str:
    """Return current time str in the format %Y-%m-%d %H:%M:%S"""
    if not for_file_name:
        cur_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
    else:
        cur_time = time.strftime("%Y-%m-%d--%H-%M-%S", time.localtime())
    return cur_time



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2024/6/25 21:10
@File    : test_config.py
@Software: PyCharm
@Desc    : 
"""
from copy import copy

import pytest

from alchemy_cat.py_tools import Config, ADict, DEP

from .test_adict import *


@pytest.fixture(scope="function")
def cfg_can_exec(dic_can_exec):
    yield Config(dic_can_exec, caps=('a', 'b'))

# -* 测试Config的文本化。


def test_cfg_to_txt(cfg_can_exec: Config):
    exec(cfg_can_exec.to_txt('exec_cfg'), globals(), ldict := {})
    exec_cfg = ldict['exec_cfg']
    assert exec_cfg == cfg_can_exec
    assert exec_cfg.uni_dic.is_whole
    assert exec_cfg.get_attribute('_cfgs_update_at_parser') == cfg_can_exec.get_attribute('_cfgs_update_at_parser')

# -* 测试初始化。


def test_not_go_in_list(dic_can_exec: ADict, cfg_can_exec: Config):
    assert cfg_can_exec.lst is dic_can_exec.lst

# TODO 一个待拆解的巨大测试。


def test_cfg_all():
    d = {'num': 0,
         'lst': [1, [2, {'lst_lst_dic': 'exit'}], {'lst_dic': 'exit'}],
         'dic': {'sub_num': 3, 'sub_set': frozenset((1, 2, 3)), 'sub_dic': {'sub_sub_tuple': (4, 5)}},
         'uni_dic': {'sub_num': 4}}

    cfg = Config(d)
    cfg.rslt_dir = '/tmp/test_alchemy_cat_config'
    cfg.uni_dic.set_whole()

    indep_tree = Config({'a': 1, 'b': 2, 'c': {'cc': 3, 'dd': {'ddd': 4}}, 'e': {'ee': 5}})
    indep_tree.c.il_rel = DEP(lambda c: c.a - c.b, rel=True)  # rel_IL
    indep_tree.c.il_not_rel = DEP(lambda c: c.num - c.i1.i2.c.dd.ddd, rel=False)  # not_rel_IL

    cfg.d1.d2.d3.k = 'v'  # missing + set_v + set_unmounted_dic

    cfg.i1.i2 = indep_tree  # set root tree

    cfg.f1.f2 = indep_tree.e  # set subtree

    new_cfg = copy(cfg)

    cfg.load(create_rslt_dir=False)

    assert cfg == {'num': 0,
                   'lst': [1, [2, {'lst_lst_dic': 'exit'}], {'lst_dic': 'exit'}],
                   'dic': {'sub_num': 3,
                           'sub_set': frozenset({1, 2, 3}),
                           'sub_dic': {'sub_sub_tuple': (4, 5)}},
                   'uni_dic': {'sub_num': 4},
                   'rslt_dir': '/tmp/test_alchemy_cat_config',
                   'd1': {'d2': {'d3': {'k': 'v'}}},
                   'i1': {'i2': {'a': 1,
                                 'b': 2,
                                 'c': {'cc': 3, 'dd': {'ddd': 4}, 'il_rel': -1, 'il_not_rel': -4},
                                 'e': {'ee': 5}}},
                   'f1': {'f2': {'ee': 5}}}

    # 均mounted且有正确父子关系。
    bs = list(cfg.branches)
    assert len(bs) == 14

    for b in bs:
        assert b._is_mounted

    root_num = 0
    for b in bs:  # 所有branch有正确的父子关系。
        if b.__parent is None:
            assert b.__key is None
            root_num += 1
        else:
            assert b.__parent[b.__key] is b
    assert root_num == 1

    # missing生成的子树链有正确的父子关系。
    assert cfg.__parent is None
    assert cfg.__key is None
    assert cfg.d1.__parent is cfg
    assert cfg.d1.__key == 'd1'
    assert cfg.d1.d2.__parent is cfg.d1
    assert cfg.d1.d2.__key == 'd2'
    assert cfg.d1.d2.d3.__parent == cfg.d1.d2
    assert cfg.d1.d2.d3.__key == 'd3'

    # 挂载独立树，能正确变更树的父子关系。
    assert cfg.i1.i2 is indep_tree
    assert cfg.i1.i2.__parent is cfg.i1
    assert cfg.i1.i2.__key == 'i2'

    # 子树、父树的内部结果不受影响。
    assert cfg.i1.i2.c.__parent is cfg.i1.i2
    assert cfg.i1.i2.c.__key == 'c'

    assert cfg.i1.__parent is cfg
    assert cfg.i1.__key == 'i1'

    # 挂载非独立树，触发拷贝。
    assert cfg.f1.f2 is not indep_tree.e
    assert cfg.f1.f2 == indep_tree.e
    assert cfg.f1.f2.__parent is cfg.f1
    assert cfg.f1.f2.__key == 'f2'

    assert indep_tree.e.__parent is indep_tree
    assert indep_tree.e.__key == 'e'

    assert cfg.f1.__parent is cfg
    assert cfg.f1.__key == 'f1'

    # 测试branch_copy
    assert len(list(cfg.branches)) == len(list(new_cfg.branches))
    for b1, b2 in zip(cfg.branches, new_cfg.branches):  # branch数目不变，但是branch的id变了。
        assert b1 is not b2

    for b in new_cfg.branches:  # 所有branch均挂载。
        assert b._is_mounted

    root_num = 0
    for b in new_cfg.branches:  # 所有branch有正确的父子关系。
        if b.__parent is None:
            assert b.__key is None
            root_num += 1
        else:
            assert b.__parent[b.__key] is b
    assert root_num == 1

    assert len(list(cfg.leaves)) == len(list(new_cfg.leaves))
    for l1, l2 in zip(cfg.leaves, new_cfg.leaves):  # 叶子不变。DEP不受影响。
        if not isinstance(l2, DEP):
            assert l1 is l2

    assert new_cfg.i1.i2.c.il_rel.level == 1
    assert new_cfg.i1.i2.c.il_not_rel.level == float('inf')

    new_cfg.load(create_rslt_dir=False)
    assert new_cfg == cfg

    # 测试删除。
    cfg.unfreeze()

    del_dic = cfg.dic
    del cfg.dic
    assert 'dic' not in cfg
    assert del_dic.__parent is None
    assert del_dic.__key is None

    pop_uni_dic = cfg.pop('uni_dic')
    assert 'uni_dic' not in cfg
    assert pop_uni_dic.__parent is None
    assert pop_uni_dic.__key is None

    clear_d1, clear_i1, clear_f1 = cfg.d1, cfg.i1, cfg.f1
    cfg.clear()
    assert len(cfg) == 0
    assert clear_d1.__parent is None
    assert clear_d1.__key is None
    assert clear_i1.__parent is None
    assert clear_i1.__key is None
    assert clear_f1.__parent is None
    assert clear_f1.__key is None

    cfg.a.b = 1
    pop_k, pop_v = cfg.popitem()
    assert len(cfg) == 0
    assert pop_k == 'a'
    assert pop_v.__parent is None
    assert pop_v.__key is None
    assert pop_v.b == 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: dataset.py
@time: 2019/12/7 23:46
@desc: A dataset lib. Quite similar to torch.util.dataset.
"""
import bisect
from typing import Union, Sequence

from numpy.random import permutation
import numpy as np
import torch
from torch.utils import data as torch_data

from alchemy_cat.alg import accumulate
from alchemy_cat.py_tools import indent


__all__ = ["Dataset", "TensorDataset", "Subset", "random_split"]


class Dataset(object):
    def __getitem__(self, index: Union[slice, list, np.ndarray, torch.Tensor, int]):
        if isinstance(index, slice):
            # Get the start, stop, and step from the slice, return an iterator
            return (self[ii] for ii in range(*index.indices(len(self))))
        if isinstance(index, list) or isinstance(index, np.ndarray) or isinstance(index, torch.Tensor):
            return [self[ii] for ii in index]
        elif isinstance(int(index), int):
            index = int(index)
            if index < 0:  # Handle negative indices
                index += len(self)
            if index < 0 or index >= len(self):
                raise IndexError("The index (%d) is out of range" % index)
            return self.get_item(index)  # Get the data from elsewhere
        else:
            raise TypeError("Invalid argument type")

    def __add__(self, other):
        return ConcatDataset([self, other])

    def get_item(self, index):
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError

    def __repr__(self):
        return f"Dataset <{self.__class__.__name__}>:\n" + \
               indent("#data: {}".format(self.__len__()))

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]


class TensorDataset(Dataset):
    r"""Dataset wrapping tensors.
    Each sample will be retrieved by indexing tensors along the first dimension.
    Arguments:
        *tensors (Tensor): tensors that have the same size of the first dimension.
    """

    def __init__(self, *tensors):
        assert all(tensors[0].size(0) == tensor.size(0) for tensor in tensors)
        self.tensors = tensors

    def get_item(self, index):
        return tuple(tensor[index] for tensor in self.tensors)

    def __len__(self):
        return self.tensors[0].size(0)


class ConcatDataset(Dataset):
    r"""Dataset as a concatenation of multiple datasets.
    This class is useful to assemble different existing datasets.
    Arguments:
        datasets (sequence): List of datasets to be concatenated
    """

    @staticmethod
    def cumsum(sequence):
        r, s = [], 0
        for e in sequence:
            l = len(e)
            r.append(l + s)
            s += l
        return r

    def __init__(self, datasets):
        super(ConcatDataset, self).__init__()
        assert len(datasets) > 0, 'datasets should not be an empty iterable'
        self.datasets = list(datasets)
        for d in self.datasets:
            assert len(d) >= 0
        self.cumulative_sizes = self.cumsum(self.datasets)

    def __len__(self):
        return self.cumulative_sizes[-1]

    def get_item(self, idx):
        if idx < 0:
            if -idx > len(self):
                raise ValueError("absolute value of index should not exceed dataset length")
            idx = len(self) + idx
        dataset_idx = bisect.bisect_right(self.cumulative_sizes, idx)
        if dataset_idx == 0:
            sample_idx = idx
        else:
            sample_idx = idx - self.cumulative_sizes[dataset_idx - 1]
        return self.datasets[dataset_idx][sample_idx]


class Subset(Dataset):
    r"""
    Subset of a dataset at specified indices.
    Arguments:
        dataset (Dataset): The whole Dataset
        indices (sequence): Indices in the whole set selected for subset
    """

    def __init__(self, dataset: Union[Dataset, torch_data.Dataset], indices: Sequence[int]):
        self.dataset = dataset
        self.indices = indices

    def get_item(self, idx):
        return self.dataset[self.indices[idx]]

    def __len__(self):
        return len(self.indices)


def random_split(dataset: Union[Dataset, torch_data.Dataset], lengths):
    r"""
    Randomly split a dataset into non-overlapping new datasets of given lengths.
    Arguments:
        dataset (Dataset): Dataset to be split
        lengths (sequence): lengths of splits to be produced
    """
    if sum(lengths) != len(dataset):
        raise ValueError("Sum of input lengths does not equal the length of the input dataset!")

    indices = permutation(sum(lengths)).tolist()
    return [Subset(dataset, indices[offset - length:offset]) for offset, length in zip(accumulate(lengths), lengths)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/3/23 11:52
@File    : ASL.py
@Software: PyCharm
@Desc    : 拷贝修改自 https://github.com/Alibaba-MIIL/ASL/blob/main/src/loss_functions/losses.py。
"""
import torch
import torch.nn as nn

__all__ = ['AsymmetricLoss', 'AsymmetricLossOptimized', 'ASLSingleLabel']


class AsymmetricLoss(nn.Module):
    def __init__(self, gamma_neg=4, gamma_pos=1, clip=0.05, eps=1e-6, disable_torch_grad_focal_loss=True,
                 reduce: str='mean'):
        super(AsymmetricLoss, self).__init__()

        self.gamma_neg = gamma_neg
        self.gamma_pos = gamma_pos
        self.clip = clip
        self.disable_torch_grad_focal_loss = disable_torch_grad_focal_loss
        self.eps = eps
        self.reduce = reduce

    def forward(self, x, y):
        """
        Parameters
        ----------
        x: input logits
        y: targets (multi-label binarized vector)
        """

        # Calculating Probabilities
        x_sigmoid = torch.sigmoid(x)
        xs_pos = x_sigmoid
        xs_neg = 1 - x_sigmoid

        # Asymmetric Clipping
        if self.clip is not None and self.clip > 0:
            xs_neg = (xs_neg + self.clip).clamp(max=1)

        # Basic CE calculation
        los_pos = y * torch.log(xs_pos.clamp(min=self.eps))
        los_neg = (1 - y) * torch.log(xs_neg.clamp(min=self.eps))
        loss = los_pos + los_neg

        # Asymmetric Focusing
        if self.gamma_neg > 0 or self.gamma_pos > 0:
            if self.disable_torch_grad_focal_loss:
                torch.set_grad_enabled(False)
            pt0 = xs_pos * y
            pt1 = xs_neg * (1 - y)  # pt = p if t > 0 else 1-p
            pt = pt0 + pt1
            one_sided_gamma = self.gamma_pos * y + self.gamma_neg * (1 - y)
            one_sided_w = torch.pow(1 - pt, one_sided_gamma)
            if self.disable_torch_grad_focal_loss:
                torch.set_grad_enabled(True)
            loss *= one_sided_w

        match self.reduce:
            case 'sum':
                return -loss.sum()
            case 'mean':
                return -loss.mean()
            case _:
                raise ValueError(f"不支持的{self.reduce=}")


class AsymmetricLossOptimized(nn.Module):
    """ Notice - optimized version, minimizes memory allocation and gpu uploading,
    favors inplace operations"""

    def __init__(self, gamma_neg=4, gamma_pos=1, clip=0.05, eps=1e-6, disable_torch_grad_focal_loss=False,
                 reduce: str='mean'):
        super(AsymmetricLossOptimized, self).__init__()

        self.gamma_neg = gamma_neg
        self.gamma_pos = gamma_pos
        self.clip = clip
        self.disable_torch_grad_focal_loss = disable_torch_grad_focal_loss
        self.eps = eps
        self.reduce = reduce

        # prevent memory allocation and gpu uploading every iteration, and encourages inplace operations
        self.targets = self.anti_targets = self.xs_pos = self.xs_neg = self.asymmetric_w = self.loss = None

    def forward(self, x, y):
        """"
        Parameters
        ----------
        x: input logits
        y: targets (multi-label binarized vector)
        """

        self.targets = y
        self.anti_targets = 1 - y

        # Calculating Probabilities
        self.xs_pos = torch.sigmoid(x)
        self.xs_neg = 1.0 - self.xs_pos

        # Asymmetric Clipping
        if self.clip is not None and self.clip > 0:
            self.xs_neg.add_(self.clip).clamp_(max=1)

        # Basic CE calculation
        self.loss = self.targets * torch.log(self.xs_pos.clamp(min=self.eps))
        self.loss.add_(self.anti_targets * torch.log(self.xs_neg.clamp(min=self.eps)))

        # Asymmetric Focusing
        if self.gamma_neg > 0 or self.gamma_pos > 0:
            if self.disable_torch_grad_focal_loss:
                torch.set_grad_enabled(False)
            self.xs_pos = self.xs_pos * self.targets
            self.xs_neg = self.xs_neg * self.anti_targets
            self.asymmetric_w = torch.pow(1 - self.xs_pos - self.xs_neg,
                                          self.gamma_pos * self.targets + self.gamma_neg * self.anti_targets)
            if self.disable_torch_grad_focal_loss:
                torch.set_grad_enabled(True)
            self.loss *= self.asymmetric_w

        match self.reduce:
            case 'sum':
                return -self.loss.sum()
            case 'mean':
                return -self.loss.mean()
            case _:
                raise ValueError(f"不支持的{self.reduce=}")


class ASLSingleLabel(nn.Module):
    """
    This loss is intended for single-label classification problems
    """
    def __init__(self, gamma_pos=0, gamma_neg=4, eps: float = 0.1, reduction='mean'):
        super(ASLSingleLabel, self).__init__()

        self.eps = eps
        self.logsoftmax = nn.LogSoftmax(dim=-1)
        self.targets_classes = []
        self.gamma_pos = gamma_pos
        self.gamma_neg = gamma_neg
        self.reduction = reduction

    def forward(self, inputs, target):
        """
        "input" dimensions: - (batch_size,number_classes)
        "target" dimensions: - (batch_size)
        """
        num_classes = inputs.size()[-1]
        log_preds = self.logsoftmax(inputs)
        self.targets_classes = torch.zeros_like(inputs).scatter_(1, target.long().unsqueeze(1), 1)

        # ASL weights
        targets = self.targets_classes
        anti_targets = 1 - targets
        xs_pos = torch.exp(log_preds)
        xs_neg = 1 - xs_pos
        xs_pos = xs_pos * targets
        xs_neg = xs_neg * anti_targets
        asymmetric_w = torch.pow(1 - xs_pos - xs_neg,
                                 self.gamma_pos * targets + self.gamma_neg * anti_targets)
        log_preds = log_preds * asymmetric_w

        if self.eps > 0:  # label smoothing
            self.targets_classes = self.targets_classes.mul(1 - self.eps).add(self.eps / num_classes)

        # loss calculation
        loss = - self.targets_classes.mul(log_preds)

        loss = loss.sum(dim=-1)
        if self.reduction == 'mean':
            loss = loss.mean()

        return loss



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2023/5/23 22:17
@File    : cache_dir.py
@Software: PyCharm
@Desc    : 
"""
import os
import os.path as osp
import subprocess
import uuid

from .str_formatters import get_local_time_str

__all__ = ['CacheDir']


class CacheDir(os.PathLike):

    def __init__(self,
                 save_at: str | os.PathLike='/dev/null',
                 cache_dir: str | os.PathLike='/tmp/cache_dir',
                 exist: str='error',
                 save_when_del: bool=False,
                 save_when_exit: bool=False,
                 enabled: bool=True):
        """将save_at下的文件缓存到cache_at下。

        Args:
            save_at: 最终保存目录。
            cache_dir: 缓存目录的父目录，推荐/tmp/xxx、/shm/xxx。
            exist: 'error' - 如果save_at已经存在，则报错；'cache': 将save_at转移到cache_at；'delete': 删除save_at。
            save_when_del: 是否在CacheDir被del时dump。
            save_when_exit: 是否在程序退出时dump。
            enabled: 是否启用CacheDir，如果不启用，初始化后，就处于_saved状态。此时等价于save_at。
        """
        super().__init__()
        save_at, cache_dir = str(save_at), str(cache_dir)

        self.save_dir = osp.dirname(save_at)
        self.save_at = save_at

        self.cache_dir = cache_dir
        self.cache_at = osp.join(self.cache_dir, f'{get_local_time_str(for_file_name=True)}@{uuid.uuid4().hex}')

        assert exist in ['error', 'cache', 'delete']
        self.exist = exist

        self.save_when_del = save_when_del
        self.save_when_exit = save_when_exit
        if self.save_when_exit:
            raise NotImplementedError(f"{type(self)}: dump_when_exit功能尚未实现。")

        self.enabled = enabled

        if enabled:
            self._saved: bool | str = False
            self._cache()
        else:
            os.makedirs(self.save_at, exist_ok=True)
            self._saved: bool | str = '未启用'

    @property
    def saved(self):
        return self._saved

    @property
    def at(self):
        if self.saved:
            return self.save_at
        else:
            return self.cache_at

    def __str__(self):
        return self.at

    def __fspath__(self):
        return self.at

    def __repr__(self):
        return f"{type(self)}: (exist: {self.exist}, saved: {self.saved})\n" \
               f"   save_at: {self.save_at} \n" \
               f"   cache_at: {self.cache_at} \n"

    def _cache(self):
        """将save_at下的文件缓存到cache_at下"""
        if self._saved:
            raise RuntimeError(f"{type(self)}: CacheDir状态为{self.saved}，不能再次cache。")

        # * 新建cache_at。
        os.makedirs(self.cache_at, exist_ok=False)

        # * 若save_at存在，则根据exist进行处理。
        if osp.isdir(self.save_at):
            match self.exist:
                case 'error':
                    raise RuntimeError(f"{type(self)}: {self.save_at} 已经存在，不能cache。")
                case 'cache':
                    print(f"{type(self)}: {self.save_at} 已经存在，正在将其转移到 {self.cache_at} ...")
                    subprocess.run(['cp', '-a', osp.join(self.save_at, ''), self.cache_at])
                case 'delete':
                    print(f"{type(self)}: {self.save_at} 已经存在，正在删除...")
                    subprocess.run(['rm', '-r', self.save_at])
                case _:
                    raise RuntimeError(f"{type(self)}: 不支持的exist参数。")

    def flush(self, size_only: bool=False):
        """将cache_at下的文件同步到save_at下

        Args:
            size_only: rsync时，只比较size。
        """
        if self._saved:
            print(f"{type(self)}:  CacheDir状态为{self.saved}，flush无效。")
            return

        if self.save_at == '/dev/null':
            print(f"{type(self)}: save_at为/dev/null，flush无效。")
            return

        # * 确保save_at的目录存在。
        os.makedirs(self.save_at, exist_ok=True)

        # * 将cache_at用rsync同步到save_at。
        print(f"{type(self)}: 正在将 {self.cache_at} 同步到 {self.save_at} ...")
        if size_only:
            subprocess.run(['rsync', '-a', '--size-only', '--delete', osp.join(self.cache_at, ''), self.save_at])
        else:
            subprocess.run(['rsync', '-a', '--delete', osp.join(self.cache_at, ''), self.save_at])

    def _rm_cache(self):
        print(f"{type(self)}: 正在删除 {self.cache_at} ...")
        subprocess.run(['rm', '-r', self.cache_at])

    def save(self, size_only: bool=False, no_flush: bool=False):
        """将cache_at下的文件转移到save_at下

        Args:
            size_only: rsync时，只比较size。
            no_flush: 如果能肯定cache和save已经一致，无需进行flush，可以设置no_flush=True。
        """
        if self._saved:
            print(f"{type(self)}: CacheDir状态为{self.saved}，save无效。")
            return

        print(f"{type(self)}: 正在将 {self.cache_at} 保存到 {self.save_at} ...")

        # * 进行最后一次同步。
        if not no_flush:
            self.flush(size_only=size_only)

        # * 删除cache_at。
        self._rm_cache()

        self._saved = '已保存'

    def terminate(self):
        """直接退出，删除cache_at。"""
        if self._saved:
            print(f"{type(self)}: CacheDir状态为{self.saved}，terminate无效。")
            return

        print(f"{type(self)}: 正在终止...")

        # * 删除cache_at。
        self._rm_cache()

        self._saved = '已终止'

    def __del__(self):
        if self.save_when_del:
            self.save()



# -- configs/mnist/plain_usage,2xbs/cfg.py --

from alchemy_cat.dl_config import Config

cfg = Config(caps='configs/mnist/plain_usage/cfg.py')

cfg.loader.ini.batch_size = 128 * 2

cfg.opt.ini.lr = 0.01 * 2



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Xiaobo Yang
@contact: hal_42@zju.edu.cn
@software: PyCharm
@file: func_tools.py
@time: 2020/1/15 1:46
@desc:
"""
__all__ = ["Lambda", "Compose", "PackCompose"]


# Copy from torch.vision
class Lambda(object):
    """Apply a user-defined lambda as a transform.

    Args:
        lambd (function): Lambda/function to be used for transform.
    """

    def __init__(self, lambd):
        assert callable(lambd), repr(type(lambd).__name__) + " object is not callable"
        self.lambd = lambd

    def __call__(self, img):
        return self.lambd(img)

    def __repr__(self):
        return self.__class__.__name__ + '()'


# Copy from torch.vision
class Compose(object):
    """Composes several transforms together.

    Args:
        transforms (list of ``Transform`` objects): list of transforms to compose.

    Example:
        >>> transforms.Compose([
        >>>     transforms.CenterCrop(10),
        >>>     transforms.ToTensor(),
        >>> ])
    """

    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, img):
        for t in self.transforms:
            img = t(img)
        return img

    def __repr__(self):
        format_string = self.__class__.__name__ + '('
        for t in self.transforms:
            format_string += '\n'
            format_string += '    {0}'.format(t)
        format_string += '\n)'
        return format_string


class PackCompose(Compose):
    """串联一系列输入、输出是多个值的函数可调用对象"""
    def __call__(self, *x):
        for t in self.transforms:
            x = t(*x)
        return x



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Author  : Xiaobo Yang
@Contact : hal_42@zju.edu.cn
@Time    : 2022/4/30 17:26
@File    : dist_communication.py
@Software: PyCharm
@Desc    : 跨卡通信工具包。
"""
import torch
import torch.distributed as dist

__all__ = ['all_gather_tensor', 'all_gather_tensor_grad']


def all_gather_tensor(tensor: torch.Tensor) -> torch.Tensor:
    """在多个rank上，将(N, ...)的张量收集到一起，拼接为(rank_num * N,...)的Tensor。梯度不会被保留。

    Args:
        tensor: (N, ...)张量。

    Returns:
        (rank_num * N,...)张量。
    """
    with torch.no_grad():  # 或冗余，强调是无梯度操作。
        tensor_list = [torch.empty_like(tensor) for _ in range(dist.get_world_size())]
        dist.all_gather(tensor_list, tensor)
    return torch.cat(tensor_list, dim=0)


def all_gather_tensor_grad(tensor: torch.Tensor) -> torch.Tensor:
    """在多个rank上，将(N, ...)的张量收集到一起，拼接为(rank_num * N,...)的Tensor。来自本rank张量的梯度会被保留。

    Args:
        tensor: (N, ...)张量。

    Returns:
        (rank_num * N, ...)张量，来自本rank张量的梯度会被保留。
    """
    with torch.no_grad():
        tensor_list = [torch.empty_like(tensor) for _ in range(dist.get_world_size())]
        dist.all_gather(tensor_list, tensor)
    tensor_list[dist.get_rank()] = tensor
    return torch.cat(tensor_list, dim=0)
